/******************************************************************************/
/*                                                                            */
/*                       X r d O s s C a c h e C o n f i g . c c              */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

//          $Id: XrdOssCacheConfig.cc,v 1.0 2009-10-19 00:00:00 ajp Exp $

const char *XrdOssCacheConfigCVSID = "$Id: XrdOssCacheConfig.cc,v 106 2009-10-19 00:00:00 ajp Exp $";

#include <unistd.h>
#include <ctype.h>
#include <strings.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <fcntl.h>

#include "XrdOssCache/XrdOssCachePlugin.hh"
#include "XrdOuc/XrdOuca2x.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdSys/XrdSysError.hh"
#include "XrdSys/XrdSysHeaders.hh"
#include "XrdSys/XrdSysPlatform.hh"
#include "XrdOuc/XrdOucStream.hh"
#include "XrdOuc/XrdOucTList.hh"
#include "XrdOuc/XrdOucUtils.hh"
#include "XrdOuc/XrdOucString.hh"
#include "XrdNet/XrdNetDNS.hh"


/******************************************************************************/
/*                               d e f i n e s                                */
/******************************************************************************/

#define TS_Xeq(x,m)    if (!strcmp(x,var)) return m(&OssCacheEroute, Config, var);

/******************************************************************************/
/*                               G l o b a l s                                */
/******************************************************************************/

char        *XrdOssCacheSys::ConfigFN;       // -> Pointer to the config file name
char        *XrdOssCacheSys::myHost;
char        *XrdOssCacheSys::myName;
long         XrdOssCacheSys::PageSize     = 4*1024;           // 4k pages
long         XrdOssCacheSys::Size         = 1024*1024*1024;   // 1G cache size
long         XrdOssCacheSys::Objects      = 1000;             // 1000 Files 
float        XrdOssCacheSys::Threshold    = 0.1;              // 10% margin
char        *XrdOssCacheSys::Directory    = (char*)"/tmp/osscache/"; // cache mount point
long         XrdOssCacheSys::SyncInterval = 10;               // every 10 seconds
bool         XrdOssCacheSys::Public       = false;            // cache objects are private
long long    XrdOssCacheSys::MapSize      = 4*1024;           // 4k memory mapped per file
bool         XrdOssCacheSys::NamespaceLink = true;            // maintain namespace links
bool         XrdOssCacheSys::WriteSync    = false;            // write's are not immedeatly synced by default
XrdOucHash<XrdOucString> *XrdOssCacheSys::delaystore;         // keeping track's how long clients have been stalled since during last attemp 
XrdOucHash<XrdOucEnv> XrdOssCacheSys::ProxyMap;

namespace XrdOssCachePlugin
{
  extern XrdSysError OssCacheEroute;
  extern XrdOucTrace OssCacheTrace;
}

using namespace XrdOssCachePlugin;

/******************************************************************************/
/*                             C o n f i g u r e                              */
/******************************************************************************/
  
int XrdOssCacheSys::Configure(const char *cfn)
{
/*
  Function: Establish configuration at start up time.

  Input:    None.

  Output:   0 upon success or !0 otherwise.
*/
   int NoGo = 0;

// Preset tracing options
//
   myHost = getenv("XRDHOST");
   myName = getenv("XRDNAME");
   if (!myName || !*myName) myName = (char *)"anon";

// Process the configuration file
//
   if ((NoGo = ConfigProc(cfn))) return NoGo;

   delaystore = new XrdOucHash<XrdOucString> ();

   if (!delaystore) {
     return 1;
   }

   return 0;
}

/******************************************************************************/
/*                     P r i v a t e   F u n c t i o n s                      */
/******************************************************************************/

/******************************************************************************/
/*                            C o n f i g P r o c                             */
/******************************************************************************/
  
int XrdOssCacheSys::ConfigProc(const char *Cfn)
{
  char *var;
  int  cfgFD, retc, NoGo = 0;
  XrdOucEnv myEnv;
  XrdOucStream Config(&OssCacheEroute, getenv("XRDINSTANCE"), &myEnv, "=====> ");

// Make sure we have a config file
//
   if (!Cfn || !*Cfn)
      {OssCacheEroute.Emsg("Config", "osscache configuration file not specified.");
       return 1;
      }

// Try to open the configuration file.
//
   if ( (cfgFD = open(Cfn, O_RDONLY, 0)) < 0)
      {OssCacheEroute.Emsg("Config", errno, "open config file", Cfn);
       return 1;
      }
   Config.Attach(cfgFD);

// Now start reading records until eof.
//
   while((var = Config.GetMyFirstWord()))
        {if (!strncmp(var, "osscache.", 9)
         ||  !strcmp(var, "all.manager")
         ||  !strcmp(var, "all.adminpath"))
            if (ConfigXeq(var+9, Config)) {Config.Echo(); NoGo = 1;}
        }

// Now check if any errors occured during file i/o
//
   if ((retc = Config.LastError()))
      NoGo = OssCacheEroute.Emsg("Config", retc, "read config file", Cfn);
   Config.Close();
   XrdOucString outval ="";
   outval += (int)(PageSize*1.0/1024.0); outval += " Kb";

   OssCacheEroute.Say("=====> osscache.pagesize      = ",outval.c_str() );
   outval ="";
   outval += (int)(Size*1.0/1024.0/1024.0); outval += " Mb";
   OssCacheEroute.Say("=====> osscache.size          = ",outval.c_str() );

   outval ="";
   outval += (int)(Objects); outval += " (max. files in cache)";
   OssCacheEroute.Say("=====> osscache.objects       = ",outval.c_str() );

   char fval[1024];
   sprintf(fval,"%.02f",Threshold);
   outval = fval; outval += " ("; outval += int(Threshold*100); outval += "% )";
   OssCacheEroute.Say("=====> osscache.threshold     = ",outval.c_str() );
   
   outval ="";
   outval += (int)(SyncInterval); outval += " (async interval between cache disk syncs in seconds)";
   OssCacheEroute.Say("=====> osscache.syncinterval  = ",outval.c_str() );

   outval ="";
   outval += (int)(MapSize); outval += " (max. memory mapped region per file)";
   OssCacheEroute.Say("=====> osscache.mapsize       = ",outval.c_str() );

   outval ="";
   outval += (int)(Public); outval += " (1:=files are world readable 0:=files are user private)";
   OssCacheEroute.Say("=====> osscache.public        = ",outval.c_str() );

   outval ="";
   outval += (int)(WriteSync); outval += " (1:=writes are sync 0:=writes are async)";
   OssCacheEroute.Say("=====> osscache.sync          = ",outval.c_str() );

   OssCacheEroute.Say("=====> osscache.directory     = ",Directory);
   
   if (access(Directory,W_OK)) {
     OssCacheEroute.Emsg("Config", errno, "access oss cache directory", Directory);
     return 1;
   }
   
// Return final return code
//
   return NoGo;
}

/******************************************************************************/
/*                             C o n f i g X e q                              */
/******************************************************************************/

int XrdOssCacheSys::ConfigXeq(char *var, XrdOucStream &Config)
{

   // Process items. for either a local or a remote configuration
   //
   TS_Xeq("trace",         xtrac);
   TS_Xeq("pagesize",      xcache);
   TS_Xeq("size",          xcache);
   TS_Xeq("objects",       xcache);
   TS_Xeq("threshold",     xcache);
   TS_Xeq("directory",     xcache);
   TS_Xeq("syncinterval",  xcache);
   TS_Xeq("public",        xcache);
   TS_Xeq("mapsize",       xcache);
   TS_Xeq("map",           xcache);
   TS_Xeq("namesapce",     xcache);
   TS_Xeq("sync",     xcache);

   // No match found, complain.
   //
   OssCacheEroute.Say("Config warning: ignoring unknown directive '",var,"'.");
   Config.Echo();
   return 0;
}
  
  
/******************************************************************************/
/*                                x t r a c e                                 */
/******************************************************************************/

/* Function: xtrace

   Purpose:  To parse the directive: trace <events>

             <events> the blank separated list of events to trace. Trace
                      directives are cummalative.

   Output: retc upon success or -EINVAL upon failure.
*/

int XrdOssCacheSys::xtrac(XrdSysError *Eroute, XrdOucStream &Config, const char* var)
{
    char  *val;
    static struct traceopts {const char *opname; int opval;} tropts[] =
       {
        {"all",      TRACE_ALL},
        {"debug",    TRACE_Debug},
        {"open",     TRACE_Open},
        {"read",      TRACE_Read},
	{"sync",      TRACE_Sync},
	{"cache",     TRACE_Cache},
	{"stats",     TRACE_Stats},
       };
    int i, trval = 0, numopts = sizeof(tropts)/sizeof(struct traceopts);

    if (!(val = Config.GetWord()))
       {Eroute->Emsg("config", "trace option not specified"); return 1;}
    while (val)
         {if (!strcmp(val, "off")) trval = 0;
             else {for (i = 0; i < numopts; i++)
                       {if (!strcmp(val, tropts[i].opname))
                           {trval |=  tropts[i].opval;
                            break;
                           }
                       }
                   if (i >= numopts)
                      Eroute->Say("Config warning: ignoring invalid trace option '",val,"'.");
                  }
          val = Config.GetWord();
         }
    OssCacheTrace.What = trval;
    // we always trace errors
    OssCacheTrace.What |= TRACE_Error;
    return 0;
}


/******************************************************************************/
/*                                x c a c h e                                 */
/******************************************************************************/

/* Function: xcache

   Purpose:  To parse cache directives: pagesize <pagesize>
             <pagesize> number of bytes of a smallest cache uniq = page size

                                        size <size>
             <size> number of bytes of the full cache space

                                        objects <n>
             <n> maximum number of files inside the cache

                                        threshold <threshold>
             <threshold> % of space to keep free in the cache (0.1 = 10%)

                                        directory <dir>
             <dir> path where the cache files are stored

                                        syncinterval <seconds>
             <seconds> number of seconds between file syncs of cache files

                                        public <pub>
             <pub> bool value making files public (ugo+r) or private (go-r)
                                        
	                                map <path> <opaque>
	     <path> is a logical path which has to be mapped to <opaque>
             <opaque> is an '&" separated list of key value pairs.
             The essential key/value pair is "rurl=root://<host>//<rprefix>"

   Output: retc upon success or -EINVAL upon failure.
*/

int XrdOssCacheSys::xcache(XrdSysError *Eroute, XrdOucStream &Config, const char* var)
{
    char  *val;
    
    if (!strcmp(var,"size")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "cache size not specified"); return 1;}
      else 
	Size = strtoll(val,0,0);
      
    }

    if (!strcmp(var,"pagesize")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "page size not specified"); return 1;}
      else 
	PageSize = strtoll(val,0,0);
        PageSize -= (PageSize%sysconf(_SC_PAGESIZE));
    }

    if (!strcmp(var,"objects")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "object number not specified"); return 1;}
      else 
	Objects = strtol(val,0,0);
    }

    if (!strcmp(var,"threshold")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "threshold not specified"); return 1;}
      else 
	Threshold = atof(val);
    }

    if (!strcmp(var,"directory")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "cache directory not specified"); return 1;}
      else 
	Directory = strdup(val);
    }

    if (!strcmp(var,"syncinterval")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "syncinterval not specified"); return 1;}
      else 
	SyncInterval = atoi(val);
    }

    if (!strcmp(var,"mapsize")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "mapsize not specified"); return 1;}
      else 
	MapSize = strtoll(val,0,10);
      // align to 
      MapSize -= (MapSize%sysconf(_SC_PAGESIZE));
    }

    if (!strcmp(var,"public")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "public (1 or 0) not specified"); return 1;}
      else {
	if (!strcmp("yes",val)) {
	  Public = true;
	} else {
	  if (!strcmp("no",val)) {
	    Public= 0;
	  } else {
	    if (!strcmp("true",val)) {
	      Public=1;
	    } else {
	      if (!strcmp("false",val)) {
		Public=0;
	      } else {
		if (!strcmp("1",val)) {
		  Public=1;
		} else {
		  if (!strcmp("0",val)) {
		    Public=0;
		  }
		}
	      }
	    }
	  }
	}
      }
    }


    if (!strcmp(var,"sync")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "sync (1 or 0) not specified"); return 1;}
      else {
	if (!strcmp("yes",val)) {
	  WriteSync = 1;
	} else {
	  if (!strcmp("no",val)) {
	    WriteSync = 0;
	  } else {
	    if (!strcmp("true",val)) {
	      WriteSync=1;
	    } else {
	      if (!strcmp("false",val)) {
		WriteSync=0;
	      } else {
		if (!strcmp("1",val)) {
		  WriteSync=1;
		} else {
		  if (!strcmp("0",val)) {
		    WriteSync=0;
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    if (!strcmp(var,"namespace")) {
      if (!(val = Config.GetWord()))
	{Eroute->Emsg("config", "namespace (1 or 0) not specified"); return 1;}
      else {
	if (!strcmp("yes",val)) {
	  NamespaceLink = true;
	} else {
	  if (!strcmp("no",val)) {
	    NamespaceLink= 0;
	  } else {
	    if (!strcmp("true",val)) {
	      NamespaceLink= true;
	    } else {
	      if (!strcmp("false",val)) {
		NamespaceLink=0;
	      } else { 
		if (!strcmp("1",val)) {
		  Public=1;
		} else {
		  if (!strcmp("0",val)) {
		    Public=0;
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    
    if (!strcmp(var,"map")) {
      XrdOucString pathmap;
      
      if (!(val = Config.GetWord())) 
	{Eroute->Emsg("config","mapL path not specified"); return 1;}
      else {
	pathmap = val;
	if (!(val = Config.GetWord())) 
	  {Eroute->Emsg("config","map: no opaque targets specified"); return 1;}
	else {
	  if (!pathmap.endswith("/")) pathmap += "/";
	  XrdOucEnv*  mapenv = new XrdOucEnv(val);
	  ProxyMap.Add(pathmap.c_str(), mapenv);
	  //	  XrdOucString outval = pathmap; outval += " => ";
	  //	  outval += val; 
	  //	  OssCacheEroute.Say("=====> osscache.map        = ",outval.c_str() );
	}
      }
    }
    return 0;
}
