/******************************************************************************/
/*                                                                            */
/*                             X r d O s s C a c h e . c c                    */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*       All Rights Reserved. See XrdInfo.cc for complete License Terms       */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*                DE-AC03-76-SFO0515 with the Deprtment of Energy             */
/******************************************************************************/

//         $Id: XrdOssCache.cc,v 1.0 2009-10-19 00:00:00 ajp Exp $

const char *XrdOssCachePluginCVSID = "$Id: XrdOssCache.cc,v 1.0 2009-10-19 00:00:00 ajp Exp $";

/******************************************************************************/
/*                             I n c l u d e s                                */
/******************************************************************************/
  
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <strings.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>

#ifdef __solaris__
#include <sys/vnode.h>
#endif

#include "XrdVersion.hh"

#include "XrdOssCache/XrdOssCachePlugin.hh"
#include "XrdPosix/XrdPosixXrootd.hh"

#include "XrdOss/XrdOssError.hh"
#include "XrdOuc/XrdOucEnv.hh"
#include "XrdSys/XrdSysError.hh"
#include "XrdSys/XrdSysHeaders.hh"
#include "XrdSys/XrdSysPlatform.hh"


/**************************************************/
/* defines ENV for client authentication          */
/**************************************************/

#define CLIENTENV(x) {                             \
     char krb5export[16384];                       \
     char sslexport[16384];                        \
     sprintf(krb5export,"FILE:/tmp/krb5cc_%d",x);  \
     sprintf(sslexport,"/tmp/proxy/x509up_u%d",x);       \
     chown(sslexport,x,0);                         \
     chown(krb5export,x,0);                        \
     setenv("KRB5CCNAME",krb5export,1);            \
     setenv("XrdSecSSLSESSION","0",1);             \
     setenv("XrdSecSSLUSERKEY",sslexport,1);	   \
     setenv("XrdSecSSLUSERCERT",sslexport,1);	   \
     setenv("XrdClientEUSER","root",1);            \
    }


/**************************************************/

/******************************************************************************/
/*                               G l o b a l s                                */
/******************************************************************************/



namespace XrdOssCachePlugin
{
  static XrdOssCacheSys   XrdOssCacheSS;
  
  XrdSysError OssCacheEroute(0, "osscache_");
  XrdOucTrace OssCacheTrace(&OssCacheEroute);
  
  void *XrdOssCacheCleanup(void *osscache);

static const int   PBsz = 3072;
}

using namespace XrdOssCachePlugin;

//class XrdOssClientEnv {
//public:

//}

/******************************************************************************/
/*                XrdOssGetSS (a.k.a. XrdOssGetStorageSystem)                 */
/******************************************************************************/
  
// This function is called by the OFS layer to retrieve the Storage System
// object. We return our proxy storage system object if configuration succeeded.
//
extern "C"
{
XrdOss *XrdOssGetStorageSystem(XrdOss       *native_oss,
                               XrdSysLogger *Logger,
                               const char   *config_fn,
                               const char   *parms)
{

// Ignore the parms (we accept none for now) and call the init routine
//
   return (XrdOssCacheSS.Init(Logger, config_fn) ? 0 : (XrdOss *)&XrdOssCacheSS);
}
}


/******************************************************************************/
/*                      o o s s _ S y s   M e t h o d s                       */
/******************************************************************************/
/******************************************************************************/
/*                                  i n i t                                   */
/******************************************************************************/
  
/*
  Function: Initialize proxy subsystem

  Input:    None

  Output:   Returns zero upon success otherwise (-errno).
*/
int XrdOssCacheSys::Init(XrdSysLogger *lp, const char *configfn)
{
   int NoGo;
   const char *tmp;

// Do the herald thing
//
   OssCacheEroute.logger(lp);
   OssCacheEroute.Say("Copr.  2009 CERN/Geneva, OssCache Version " XrdVSTRING);

// Initialize the subsystems
//
   tmp = ((NoGo=Configure(configfn)) ? "failed." : "completed.");
   OssCacheEroute.Say("------ OssCache storage system initialization ", tmp);

   //   if (!NoGo) NoGo = XrdOss::Init(lp,configfn);

   XrdOssSS = this;
   OssEroute.logger(lp);

   if (!NoGo) NoGo = XrdOssSys::Configure(configfn, OssEroute);

   // create a ReadCache object
   Cache = new XrdOssReadCache(Directory,PageSize, Size, Objects, SyncInterval, Threshold, MapSize, Public);
   if ((!Cache) || (!Cache->Init())) {
     OssCacheEroute.Say("=====> OssCache cache has been disabled - cache directory is (probably) unaccessible");
     delete Cache;
     Cache = 0;
   } else {
     Cache->SetSync(WriteSync);
     if (((XrdOssCacheSys*)XrdOssSS)->NamespaceLink) {
       // run initial clean-up procedure
       XrdOucString findstring = "find "; findstring += ((XrdOssCacheSys*)XrdOssSS)->Directory; findstring += "namespace"; findstring += " -type l";
       FILE* fp = popen(findstring.c_str(),"r");
       if (fp) {
	 int item=0;
	 char lpath[16384];
	 struct stat buf;
	 while ( (item = fscanf(fp,"%s\n",lpath)) == 1) {
	   // check if we can stat the cache file
	   if (stat(lpath,&buf)) {
	     // we can't -> remove it
	     if(!unlink(lpath)) {
	       XrdOucString sayit="=====> OssCache removed from namespace dead cache file: "; sayit+= lpath;
	       OssCacheEroute.Say(sayit.c_str());
	     }
	   }
	 }
	 fclose(fp);
       }
     }
   }

// All done.
//
   return NoGo;
}


bool XrdOssCacheSys::CreateLink(XrdOssReadCacheHandle* handle, const char* path) {
  if ((!handle)||(!path)) return false;

  XrdOucString namespacename = ((XrdOssCacheSys*)XrdOssSS)->Directory; namespacename += "namespace";
  namespacename += path;
  XrdOucString namespacelink = handle->cfname;

  struct stat buf;

  // check first if we have this link already 
  if (!stat(namespacename.c_str(),&buf)) 
    return true;

  int posr=STR_NPOS;

  XrdOucString namespacesubdir;
  // backward scan
  while ( (posr = namespacename.rfind('/',posr) ) != STR_NPOS ) {
    namespacesubdir.assign(namespacename,0,posr);
    posr--;
    if (!stat(namespacesubdir.c_str(),&buf)) {
      break;
    }    
  }

  if (posr != STR_NPOS) {
    posr+=2;
  }

  // foward creation
  while ( (posr = namespacename.find('/',posr) ) != STR_NPOS ) {
    namespacesubdir.assign(namespacename,0,posr);
    if (posr!= STR_NPOS) {
      if (mkdir(namespacesubdir.c_str(),0755)) 
	return false;
    }
    posr++;
  } 

  // create link
  if (symlink(namespacelink.c_str(),namespacename.c_str())) 
    return false;

  return true;
}

/******************************************************************************/
/*                                 C h m o d                                  */
/******************************************************************************/
/*
  Function: Change file mode.

  Input:    path        - Is the fully qualified name of the target file.
            mode        - The new mode that the file is to have.

  Output:   Returns XrdOssOK upon success and -errno upon failure.

  Notes:    This function is currently unsupported.
*/

int XrdOssCacheSys::Chmod(const char *path, mode_t mode)
{
  XrdOucString MapUrl;
  if (GetMapping(MapUrl,path, 0, 1)) {
    if (Maps2Remote(MapUrl.c_str())) {
      return -EIO;
    }
  }
  return XrdOssSys::Chmod(path,mode);
}

/******************************************************************************/
/*                                c r e a t e                                 */
/******************************************************************************/

/*
  Function: Create a file named `path' with 'file_mode' access mode bits set.

  Input:    path        - The fully qualified name of the file to create.
            access_mode - The Posix access mode bits to be assigned to the file.
                          These bits correspond to the standard Unix permission
                          bits (e.g., 744 == "rwxr--r--").
            env         - Environmental information.
            opts        - Set as follows:
                          XRDOSS_mkpath - create dir path if it does not exist.
                          XRDOSS_new    - the file must not already exist.
                          x00000000     - x are standard open flags (<<8)

  Output:   Returns XrdOssOK upon success; (-errno) otherwise.

*/
int XrdOssCacheSys::Create(const char *tident, const char *path, mode_t Mode,
                        XrdOucEnv &env, int Opts)
{
  bool isremote=false;
  XrdOucEnv* mapenv=NULL;
  XrdOucString fullpath = path;
  int rpos=STR_NPOS;
  printf("Enter loop \n");
  while ((rpos = fullpath.rfind("/",rpos))!=STR_NPOS) {
    rpos--;
    printf("rpos is %d [%d]\n",rpos,STR_NPOS);
    XrdOucString subpath = fullpath;
    if (rpos>-1)
      subpath.erase(rpos+2);
    if ((mapenv = ((XrdOssCacheSys*)XrdOssSS)->ProxyMap.Find(subpath.c_str()))) {
      isremote=true;
      break;
    }
    if (rpos == STR_NPOS) 
      break;
  }
  if (isremote) {
    return XrdOssOK;
  } else {
    return XrdOssSys::Create(tident,path,Mode,env,Opts);
  }
}

/******************************************************************************/
/*                                R e n a m e                                 */
/******************************************************************************/

/*
  Function: Renames a file with name 'old_name' to 'new_name'.

  Input:    old_name  - Is the fully qualified name of the file to be renamed.
            new_name  - Is the fully qualified name that the file is to have.

  Output:   Returns XrdOssOK upon success and -errno upon failure.
*/
int XrdOssCacheSys::Rename(const char *oldname, const char *newname)
{
  return XrdOssSys::Rename(oldname,newname);
}

/******************************************************************************/
/*                              T r u n c a t e                               */
/******************************************************************************/
/*
  Function: Truncate a file.

  Input:    path        - Is the fully qualified name of the target file.
            flen        - The new size that the file is to have.

  Output:   Returns XrdOssOK upon success and -errno upon failure.
*/

int XrdOssCacheSys::Truncate(const char *path, unsigned long long flen)
{
  XrdOucString MapUrl;
  if (GetMapping(MapUrl,path, 0, 1)) {
    if (Maps2Remote(MapUrl.c_str())) {
      return -EIO;
    }
  }

  return Truncate(path,flen);
}
  
/******************************************************************************/
/*                                U n l i n k                                 */
/******************************************************************************/

/*
  Function: Delete a file from the namespace and release it's data storage.

  Input:    path      - Is the fully qualified name of the file to be removed.

  Output:   Returns XrdOssOK upon success and -errno upon failure.
*/
int XrdOssCacheSys::Unlink(const char *path, int Opts)
{
  XrdOucString MapUrl;
  if (GetMapping(MapUrl,path, 0, 1)) {
    if (Maps2Remote(MapUrl.c_str())) {
      return -EIO;
    }
  }
  return Unlink(path,Opts);
}

int XrdOssCacheSys::Stat(const char *path, struct stat *buff, int resonly) {
  XrdOucString MapUrl;
  if (GetMapping(MapUrl,path, 0, 1)) {
    if (Maps2Remote(MapUrl.c_str())) {
      /*      memset(buff,sizeof(struct stat),0);
	      buff->st_blksize= 64*1024;
	      buff->st_dev    = 0;
	      buff->st_rdev   = 0;
	      buff->st_nlink  = 1;
	      buff->st_uid    = 0;
	      buff->st_gid    = 0;
	      buff->st_size   = 0;
	      buff->st_atime  = 0;
	      buff->st_mtime  = 0;
	      buff->st_ctime  = 0;
	      buff->st_blocks = 1024;
	      buff->st_ino    = 0;
	      buff->st_mode   = S_IXUSR|S_IRUSR|S_IWUSR |S_IFREG;
	      return XrdOssOK;
      */
      return -EIO;
    }
  } else XrdOssSys::Stat(path,buff,resonly);
}

bool   
XrdOssCacheSys::Maps2Local(const char* url){
  XrdOucString surl=url;
  if (surl.beginswith("root://")) {
    return false;
  } else {
    return true;
  }
}


bool   
XrdOssCacheSys::Maps2Remote(const char* url) {
  XrdOucString surl=url;
  if (surl.beginswith("root://")) {
    return true;
  } else {
    return false;
  }
}

const char* 
XrdOssCacheSys::GetMapping(XrdOucString &MapUrl, const char* path, const char* secuser, bool iswrite) {
  XrdOucEnv* mapenv=NULL;
  XrdOucString fullpath = path;
  int rpos=STR_NPOS;
  while ((rpos = fullpath.rfind("/",rpos))!=STR_NPOS) {
    rpos--;
    XrdOucString subpath = fullpath;
    if (rpos>-1) {
      subpath.erase(rpos+2);
    }
    if ((mapenv = ((XrdOssCacheSys*)XrdOssSS)->ProxyMap.Find(subpath.c_str()))) {
      const char* mapurl = 0;
      if ( iswrite) 
	mapurl = mapenv->Get("wurl");
      if (!mapurl)
	mapurl = mapenv->Get("rurl");
      
      if ( mapurl ) {
	MapUrl = fullpath;
	MapUrl.replace(subpath.c_str(),mapurl);
	if (secuser) {
	  MapUrl.replace("USER",secuser);
	}
	return MapUrl.c_str();
      }
    }
    if (rpos==STR_NPOS)
      break;
  }
  return 0;
}

int
XrdOssCacheSys::GetDelayValue(const char* tag)
{
  XrdOucString *delayval;

  if ( (delayval=delaystore->Find(tag)) ) {
    float oldval = atoi(delayval->c_str());
    oldval *= 1.8;
    if (oldval > 3600 ) {
      // more than 1 hour, doesn't make sense
      oldval = 3600;
    }
    
    (*delayval) = "";
    *(delayval) += (int) oldval;
  } else {
    delayval = new XrdOucString();
    *delayval = 20 + (rand()% 20);
    delaystore->Add(tag,delayval,3600);
  }
  return atoi(delayval->c_str());
}

/******************************************************************************/
/*                     o o s s _ F i l e   M e t h o d s                      */
/******************************************************************************/
/******************************************************************************/
/*                                  o p e n                                   */
/******************************************************************************/

/*
  Function: Open the file `path' in the mode indicated by `Mode'.

  Input:    path      - The fully qualified name of the file to open.
            Oflag     - Standard open flags.
            Mode      - Create mode (i.e., rwx).
            env       - Environmental information.

  Output:   XrdOssOK upon success; -errno otherwise.
*/
int XrdOssCacheFile::Open(const char *path, int Oflag, mode_t Mode, XrdOucEnv &Env)
{
  EPNAME("Open");
  const char* val;
  XrdOucString rpath = path;
  rpath+= "?";
  int envlen;
  rpath+= Env.Env(envlen);
  bool iswrite=false;
  bool userwantsnocache=false;
  bool doisonline;
  bool dostaging;
  TRACE(Open,"oss file open" << path);
  if (((Oflag & O_WRONLY)) || ((Oflag & O_RDWR)) || ((Oflag & O_CREAT)) ) 
    iswrite = true;
  
  if ((val = Env.Get("osscache.nocache"))) {
    userwantsnocache = atoi(val);
  }
  
  GetMapping(path, Env.Get("sec&user"),iswrite);
  
  //////////////////////////////////////////
  // check if local or remote 
  //////////////////////////////////////////

  if ( Maps2Remote() ) {
    // if this is a file stored on remote storage we create a client object connecting to the remote storage
    EnvPutInt(NAME_CONNECTTIMEOUT,150);
    EnvPutInt(NAME_REQUESTTIMEOUT,300);
    EnvPutInt(NAME_MAXREDIRECTCOUNT,10);
    EnvPutInt(NAME_RECONNECTWAIT,150);
    EnvPutInt(NAME_FIRSTCONNECTMAXCNT,1);
    EnvPutInt(NAME_DATASERVERCONN_TTL,3600);
    //      EnvPutInt(NAME_READAHEADSTRATEGY,1);
    // check for readahead specified in mapping ...
    if ((val = GetMappingKey(path,"readaheadsize"))) {
      int raheadsize=atoi(val);
      int rcachesize= 3 * raheadsize;
      if (( val = GetMappingKey(path,"readcachesize"))) {
	rcachesize = atoi(val);
      }
      EnvPutInt(NAME_READCACHESIZE,rcachesize);
      EnvPutInt(NAME_READAHEADSIZE,raheadsize);
    } else {
      EnvPutInt(NAME_READCACHESIZE,0);
    }
    
    if ((val = Env.Get("osscache.stream"))) {
      if (atoi(val)) {
	// look additionally for streaming parameters
	if ((val = GetMappingKey(path,"streamreadaheadsize"))) {
	  int raheadsize=atoi(val);
	  int rcachesize= 3 * raheadsize;
	  if (( val = GetMappingKey(path,"streamreadcachesize"))) {
	    rcachesize = atoi(val);
	  }
	  EnvPutInt(NAME_READCACHESIZE,rcachesize);
	  EnvPutInt(NAME_READAHEADSIZE,raheadsize);
	}
      }
    }
    
    // check for multistream setting
    if ((val = GetMappingKey(path,"multistreamcnt"))) {
      int mstreamcnt = atoi(val);
      EnvPutInt(NAME_MULTISTREAMCNT,mstreamcnt);
    }

    if (( val = GetMappingKey(path,"debug")))  {   
      EnvPutInt(NAME_DEBUG,atoi(val));
    }

    if (( val = GetMappingKey(path,"secdebug"))) {
      setenv("XrdSecDEBUG",val,1);
    }
    
    // check if we should check for staging
    if (( val = GetMappingKey(path,"checkisonline"))) {
      doisonline = atoi(val);
    }

    // check if we are allowed to ask for staging
    if (( val = GetMappingKey(path,"stage"))) {
      dostaging = atoi(val);
    }
    
    // create only a new client, if it does not yet exist from previous trials
    if (!Client) {
      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.Lock();
      //      Client = new XrdClient(GetMapping(path), &ClientCallback, (void*)this); // removed asynchronous calls
      Client = new XrdClient(GetMapping(path));
      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
    }

    if (!ClientAdmin) {
      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.Lock();
      ClientAdmin = new XrdClientAdmin(GetMapping(path));
      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
    }
    
    if (!Client) return -ENOMEM;
    if (!ClientAdmin) return -ENOMEM;
  }


  if ((!(Oflag & O_WRONLY)) && (!(Oflag & O_RDWR)) && (!(Oflag & O_CREAT)) && ((XrdOssCacheSys*)XrdOssSS)->GetCache()) {
    //////////////////////////////////////////
    // read only file open via cache
    //////////////////////////////////////////
    
    TRACE(Debug,"cache in use");
    // Create a cache handle
    struct stat st;


    if (!Client) {
      //////////////////////////////////////////
      // local file
      XrdOssSS->Stat(path, &st);
      if ( (!((val = GetMappingKey(path,"nocache")))) && (!userwantsnocache) && ((XrdOssCacheSys*)XrdOssSS)->GetCache()) {
	
	// if the path has not the 'nocache' option set, we read through the cache
	ch = ((XrdOssCacheSys*)XrdOssSS)->GetCache()->GetReadCacheHandle(path, st.st_size, st.st_mtime, 0);
	
	if (ch) {
	  // maintain a cache namespace if configured
	  if (((XrdOssCacheSys*)XrdOssSS)->NamespaceLink) {
	    if (!((XrdOssCacheSys*)XrdOssSS)->CreateLink(ch, path)) {
	      // namespace creation failed
	      return -errno;
	    }
	  }
	}
      }
      return XrdOssFile::Open(path,Oflag, Mode,Env);
    } else {
      //////////////////////////////////////////
      // remote file
      Client->UseCache(true);

      // setup forward authentication variables and mutex for authentication - that is not ideal but there is no better way in the XrdClient right now

      // nobody
      uid_t uid=99;

      if (( val = Env.Get("sec&user"))) {
	// translate client name to UID
	struct passwd* passwd;
	passwd = getpwnam(val);
	if (passwd) {
	  uid = passwd->pw_uid;
	}
      }

      CLIENTENV(uid);

      TRACE(Open,"forwarding client: KRB5CCNAME=" << getenv("KRB5CCNAME") << " XrdSecSSLUSERKEY=" << getenv("XrdSecSSLUSERKEY") << " XrdSecSSLUSERCERT=" << getenv("XrdSecSSLUSERCERT"));


      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.Lock();
      // in case we should check if the file is online
      if (doisonline) {
	vecString vecpath;
	vecBool   vecbool;
	XrdClientUrlInfo u;
	u.TakeUrl(GetMapping(path));
	XrdOucString newfile = u.File.c_str();
	vecpath.Push_back(newfile);
	ClientAdmin->Connect();
	bool goterror=false;
	XrdOucString errormsg="";
	int exitcode=0;
	
	if (!ClientAdmin->IsFileOnline(vecpath, vecbool)) {
	  switch (ClientAdmin->LastServerResp()->status) {
	  case kXR_ok:
	    break;
	  case kXR_error:
	    exitcode = ClientAdmin->LastServerError()->errnum;
	    errormsg = ClientAdmin->LastServerError()->errmsg;
	    break;
	  default:
	    // this should not happen!
	    break;
	  }
	}
	
	if (exitcode != 0) {
	  ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
	  TRACE(Error,"ClientAdmin::IsFileOnline path:" << GetMapping(path) << " returned code:" << exitcode << " msg:" << errormsg);
	  return -EIO;
	}
	
	if (!vecbool[0]) {
	  // file is not online ....
	  if (dostaging) {
	    TRACE(Open,"ClientAdmin::IsFileOnline path:" << GetMapping(path) << " is offline!");
	    // send stage request
	    if (!ClientAdmin->Prepare(u.File.c_str(),(kXR_char)kXR_stage, 0)) {
	      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
	      TRACE(Error,"ClientAdmin::Prepare path:" << GetMapping(path) << " failed");
#ifdef __macos__
	      return -EREMOTE;
#else
	      return -EREMOTEIO;
#endif
	    } else {
	      TRACE(Open,"ClientAdmin::Prepare path:" << GetMapping(path) << " has been asked for stagein!");
	      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
	      // ok, back off the client
	      XrdOucString delaytag = "";
	      delaytag += (int)uid;
	      delaytag += "@";
	      delaytag += path;
	      TRACE(Open,"open in progress -> bouncing client back for <<" << BounceInterval<<"secs ...");
	      int bouncetime = ((XrdOssCacheSys*)XrdOssSS)->GetDelayValue(delaytag.c_str());
	      return bouncetime;
	    }
	  } else {
	    ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
	    TRACE(Error,"ClientAdmin::IsFileOnline path:"<< GetMapping(path) << " is offline but staging is forbidden");
	    return -EACCES;
	  }
	} else {
	  TRACE(Open,"ClientAdmin::IsFileOnline path:" << GetMapping(path) << " is online!");
	}    
      }
      
      // here we are sure, that the file is online and we run a blocking Open
      
      if (!Client->Open(0,kXR_async)) {
	((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
	return -EIO;
      }
      
      ((XrdOssCacheSys*)XrdOssSS)->ClientMutex.UnLock();
  
      //	printf("Remote size is %u %u [ Client = %d ] \n",st.st_size, st.st_mtime,Client);
      if ( (!((val = GetMappingKey(path,"nocache")))) && (!userwantsnocache)) {
	if (Fstat(&st) == XrdOssOK) {
	  // if the path has not the 'nocache' option set, we read through the cache
	  ch = ((XrdOssCacheSys*)XrdOssSS)->GetCache()->GetReadCacheHandle(path, st.st_size, st.st_mtime, 0);
	  
	  if (ch) {
	    // maintain a cache namespace if configured
	    if (((XrdOssCacheSys*)XrdOssSS)->NamespaceLink) {
	      if (!((XrdOssCacheSys*)XrdOssSS)->CreateLink(ch, path)) {
		return -errno;
	      }
	    }
	  }
	} else {
	  return -EIO;
	}
      } 
      return XrdOssOK;
    }
  } else {
    //////////////////////////////////////////
    // write file open 
    //////////////////////////////////////////
    // write a file

    if (val = GetMappingKey(path,"ronly")) {
      if (atoi(val)) {
	return -EROFS;
      }
    }
    
    if (Client) {

      int XOflags, XMode;
      XMode=0;
      // Translate option bits to the appropraite values. Always
      // make directory path for new file.
      //
      XOflags = (Oflag & (O_WRONLY | O_RDWR) ? kXR_open_updt : kXR_open_read);
	XOflags |= kXR_mkpath;
      if (Oflag & O_CREAT) {
	XOflags |= (Oflag & O_EXCL ? kXR_new : kXR_delete);
      } else if (Oflag & O_TRUNC && XOflags & kXR_open_updt) {
	XOflags |= kXR_delete;
      }

      if (Mode && (Oflag & O_CREAT)) {
	if (Mode & S_IRUSR) XMode |= kXR_ur;
	if (Mode & S_IWUSR) XMode |= kXR_uw;
	if (Mode & S_IXUSR) XMode |= kXR_ux;
	if (Mode & S_IRGRP) XMode |= kXR_gr;
	if (Mode & S_IWGRP) XMode |= kXR_gw;
	if (Mode & S_IXGRP) XMode |= kXR_gx;
	if (Mode & S_IROTH) XMode |= kXR_or;
	if (Mode & S_IXOTH) XMode |= kXR_ox;
      }

      Client->UseCache(true);

      // nobody
      uid_t uid=99;

      if (( val = Env.Get("sec&user"))) {
	// translate client name to UID
	struct passwd* passwd;
	passwd = getpwnam(val);
	if (passwd) {
	  uid = passwd->pw_uid;
	}
      }
      
      CLIENTENV(uid);

      TRACE(Open,"forwarding client: KRB5CCNAME=" << getenv("KRB5CCNAME") << " XrdSecSSLUSERKEY=" << getenv("XrdSecSSLUSERKEY") << " XrdSecSSLUSERCERT=" << getenv("XrdSecSSLUSERCERT"));

      if (!Client->Open(XMode,XOflags,0)) {
	return -EIO;
      }
      return XrdOssOK;
    } else {
      // local write
      TRACE(Open,"cache bypass write");
      return XrdOssFile::Open(path,Oflag, Mode, Env);
    }
  }
}

/******************************************************************************/
/*                                 c l o s e                                  */
/******************************************************************************/

/*
  Function: Close the file associated with this object.

  Input:    None.

  Output:   Returns XrdOssOK upon success aud -errno upon failure.
*/
int XrdOssCacheFile::Close(long long *retsz)
{
  EPNAME("close");
  if (ch) {
    //    ((XrdOssCacheSys*)XrdOssSS)->HandleMutex.Lock();

    // get statistics
    char hitrate[1024];
    sprintf(hitrate,"%.02f",100.0*(ch->pagehits)/(ch->pagehits+ch->pagemiss));
    TRACE(Stats,"Pagehits: " << ch->pagehits << " Pagemiss: " << ch->pagemiss << " Pagein Bytes: " << ch->pagedbytes << " Requested Bytes: " << ch->requestedbytes << " Hitrate: " << hitrate);
    ((XrdOssCacheSys*)XrdOssSS)->GetCache()->RemoveReadCacheHandle(ch->FileName.c_str());
    ch = 0;
    //    ((XrdOssCacheSys*)XrdOssSS)->HandleMutex.UnLock();
  }
  
  if (Client) {
    bool rc = Client->Close();
    delete Client;
    Client = 0;
    return rc;
  } else {
    return XrdOssFile::Close(retsz);
  }
}

/******************************************************************************/
/*                                  r e a d                                   */
/******************************************************************************/

/*
  Function: Preread `blen' bytes from the associated file.

  Input:    offset    - The absolute 64-bit byte offset at which to read.
            blen      - The size to preread.

  Output:   Returns zero read upon success and -errno upon failure.
*/

ssize_t XrdOssCacheFile::Read(off_t offset, size_t blen)
{
     if (fd < 0) return (ssize_t)-XRDOSS_E8004;

     return 0;  // We haven't implemented this yet!
}


/******************************************************************************/
/*                                  r e a d                                   */
/******************************************************************************/

/*
  Function: Read `blen' bytes from the associated file, placing in 'buff'
            the data and returning the actual number of bytes read.

  Input:    buff      - Address of the buffer in which to place the data.
            offset    - The absolute 64-bit byte offset at which to read.
            blen      - The size of the buffer. This is the maximum number
                        of bytes that will be read.

  Output:   Returns the number bytes read upon success and -errno upon failure.
*/

ssize_t XrdOssCacheFile::Read(void *buff, off_t offset, size_t blen)
{
  EPNAME("Read");
  TRACE(Read,blen << "@"<< offset);  
  ssize_t nr = 0;
  if (ch) {
    ReadMutex.Lock();
    nr = ch->Read(offset,(char*)buff,blen, this);
    ReadMutex.UnLock();
  } 
  if ((size_t)nr != blen) {
    if (Client) {
      nr = Client->Read(buff,offset,blen);
    } else {
      nr = XrdOssFile::Read(buff,offset,blen);
    }
    TRACE(Read,"Direct=>Read");
  } else {
    TRACE(Read,"Cache=>Read");
  }
  return nr;
}

ssize_t XrdOssCacheFile::ReadNoCache(void *buff, off_t offset, size_t blen)
{
  EPNAME("ReadNoCache");
  TRACE(Cache,"Read-Direct");  
  if (Client) {
    // cached from remote disk
    int nread = Client->Read(buff,offset,blen);
    return nread;
  } else {
    // cached from local disk
    return XrdOssFile::Read(buff,offset,blen);
  }
}

/******************************************************************************/
/*                               R e a d R a w                                */
/******************************************************************************/

/*
  Function: Read `blen' bytes from the associated file, placing in 'buff'
            the data and returning the actual number of bytes read.

  Input:    buff      - Address of the buffer in which to place the data.
            offset    - The absolute 64-bit byte offset at which to read.
            blen      - The size of the buffer. This is the maximum number
                        of bytes that will be read.

  Output:   Returns the number bytes read upon success and -errno upon failure.
*/

ssize_t XrdOssCacheFile::ReadRaw(void *buff, off_t offset, size_t blen)
{
     return Read(buff, offset, blen);
}

/******************************************************************************/
/*                                 w r i t e                                  */
/******************************************************************************/

/*
  Function: Write `blen' bytes to the associated file, from 'buff'
            and return the actual number of bytes written.

  Input:    buff      - Address of the buffer from which to get the data.
            offset    - The absolute 64-bit byte offset at which to write.
            blen      - The number of bytes to write from the buffer.

  Output:   Returns the number of bytes written upon success and -errno o/w.
*/

ssize_t XrdOssCacheFile::Write(const void *buff, off_t offset, size_t blen)
{
  if (Client) {
    return Client->Write(buff,offset,blen);
  } else {
    return XrdOssFile::Write(buff,offset,blen);
  }
}

/******************************************************************************/
/*                                 f s t a t                                  */
/******************************************************************************/

/*
  Function: Return file status for the associated file.

  Input:    buff      - Pointer to buffer to hold file status.

  Output:   Returns XrdOssOK upon success and -errno upon failure.
*/

int XrdOssCacheFile::Fstat(struct stat *buff)
{
  if (Client) {
    Client->IsOpen_wait();
    XrdClientStatInfo cstatinfo;
    bool rc = Client->Stat(&cstatinfo);
    if (rc==true) {
      int newflags = 0;
      // Map the xroot flags to unix flags
      //
      if (cstatinfo.flags & kXR_xset)     newflags |= S_IXUSR;
      if (cstatinfo.flags & kXR_readable) newflags |= S_IRUSR;
      if (cstatinfo.flags & kXR_writable) newflags |= S_IWUSR;
      if (cstatinfo.flags & kXR_other)    newflags |= S_IFBLK;
      else if (cstatinfo.flags & kXR_isDir) newflags |= S_IFDIR;
      else newflags |= S_IFREG;
      if (cstatinfo.flags & kXR_offline) newflags |= S_ISVTX;
      if (cstatinfo.flags & kXR_poscpend)newflags |= S_ISUID;


      memset(buff,sizeof(struct stat),0);
      buff->st_blksize= 64*1024;
      buff->st_dev    = 0;
      buff->st_rdev   = 0;
      buff->st_nlink  = 1;
      buff->st_uid    = 0;
      buff->st_gid    = 0;
      buff->st_size   = cstatinfo.size;
      buff->st_atime  = cstatinfo.modtime;
      buff->st_mtime  = cstatinfo.modtime;
      buff->st_ctime  = cstatinfo.modtime;
      buff->st_blocks = 1024;
      buff->st_ino    = cstatinfo.id;
      buff->st_mode   = newflags;
      return XrdOssOK;
    } else {
      return -EIO;
    }
  }
  return XrdOssFile::Fstat(buff);
}

/******************************************************************************/
/*                               f s y n c                                    */
/******************************************************************************/

/*
  Function: Synchronize associated file.

  Input:    None.

  Output:   Returns XrdOssOK upon success and -errno upon failure.
*/
int XrdOssCacheFile::Fsync(void)
{
  return XrdOssFile::Fsync();
}

/******************************************************************************/
/*                             f t r u n c a t e                              */
/******************************************************************************/

/*
  Function: Set the length of associated file to 'flen'.

  Input:    flen      - The new size of the file.

  Output:   Returns XrdOssOK upon success and -errno upon failure.

  Notes:    If 'flen' is smaller than the current size of the file, the file
            is made smaller and the data past 'flen' is discarded. If 'flen'
            is larger than the current size of the file, a hole is created
            (i.e., the file is logically extended by filling the extra bytes 
            with zeroes).

            If compiled w/o large file support, only lower 32 bits are used.
            used.

            Currently not supported for proxies.
*/
int XrdOssCacheFile::Ftruncate(unsigned long long flen)
{
  return XrdOssFile::Ftruncate(flen);
}

/******************************************************************************/
/*                               g e t M m a p                                */
/******************************************************************************/
  
/*
  Function: Indicate whether or not file is memory mapped.

  Input:    addr      - Points to an address which will receive the location
                        memory where the file is mapped. If the address is
                        null, true is returned if a mapping exist.

  Output:   Returns the size of the file if it is memory mapped (see above).
            Otherwise, zero is returned and addr is set to zero.
*/
off_t XrdOssCacheFile::getMmap(void **addr)   // Not Supported for proxies
{
   if (addr) *addr = 0;
   return 0;
}
  


/******************************************************************************/
/*                     P r i v a t e    S e c t i o n                         */
/******************************************************************************/

bool XrdOssCacheFile::Maps2Local()
{
  if (Mapped) {
    if (MapUrl.beginswith("root://")) return false;
  }
  return true;
}


bool XrdOssCacheFile::Maps2Remote() 
{
  if (Mapped) {
    if (MapUrl.beginswith("root://")) return true;
  }
  return false;
}

const char* XrdOssCacheFile::GetMapping(const char* path, const char* secuser, bool iswrite) 
{
  if (!Mapped) {
    XrdOucEnv* mapenv=NULL;
    XrdOucString fullpath = path;
    int rpos=STR_NPOS;
    while ((rpos = fullpath.rfind("/",rpos))!=STR_NPOS) {
      rpos--;
      XrdOucString subpath = fullpath;
      if (rpos>-1)
	subpath.erase(rpos+2);
      if ((mapenv = ((XrdOssCacheSys*)XrdOssSS)->ProxyMap.Find(subpath.c_str()))) {
	const char* mapurl = 0;
	if ( iswrite) 
	  mapurl = mapenv->Get("wurl");
	if (!mapurl)
	  mapurl = mapenv->Get("rurl");
	
	if ( mapurl ) {
	  printf("path = %s\n",path);
	  MapUrl = fullpath;
	  MapUrl.replace(subpath.c_str(),mapurl);
	  if (secuser) {
	    MapUrl.replace("USER",secuser);
	  }
	  Mapped=true;
	  return MapUrl.c_str();
	}
      }
      if (rpos==STR_NPOS)
	break;
    }
    return 0;
  } else {
    return MapUrl.c_str();
  }
  return 0;
}

const char* XrdOssCacheFile::GetMappingKey(const char* path, const char* key) 
{
  XrdOucEnv* mapenv=NULL;
  XrdOucString fullpath = path;
  int rpos=STR_NPOS;
  while ((rpos = fullpath.rfind("/",rpos))!=STR_NPOS) {
    rpos--;
    XrdOucString subpath = fullpath;
    subpath.erase(rpos+2);
    if ((mapenv = ((XrdOssCacheSys*)XrdOssSS)->ProxyMap.Find(subpath.c_str()))) {
      const char* mapkey;
      if ( mapkey = mapenv->Get(key) ) {
	return mapkey;
      }
    }
    if (rpos==STR_NPOS)
      break;
  }
  return 0;
}
