#include <XrdOssCache/XrdOssHttpRead.hh>

int
XrdOssHttpRead::mempref(const char * mem, const char * pref, size_t size)
{
  /* return true if found */
  if (size < strlen(pref)) return 0;
  return ! memcmp(mem, pref, strlen(pref));
}


int 
XrdOssHttpRead::parse_url(const char * url, struct_url* res)
{
  const char * url_orig = url;
  const char * http = "http://";
  int path_start = '/';
  
  if (strncmp(http, url, strlen(http)) == 0) {
    url += strlen(http);
    res->proto = PROTO_HTTP;
    res->port = 80; 
  } else {
    fprintf(stderr, "Invalid protocol in url: %s\n", url_orig);
    return -1;
  }
  
  /* determine if path was given */
  if(strchr(url, path_start))
    res->path = strdup(strchr(url, path_start));
  else{
    path_start = 0;
    res->path = strdup("/");
  }
  
  
  /* Get port number. */
  int host_end = path_start;
  if(strchr(url, ':') && (strchr(url, ':') < strchr(url, path_start))){
    /* FIXME check that port is a valid numeric value */
    res->port = atoi(strchr(url, ':') + 1);
    if (! res->port) {
      fprintf(stderr, "Invalid port in url: %s\n", url_orig);
      return -1;
    }
    host_end = ':';
  }
  /* Get the host name. */
  if (url == strchr(url, host_end)){ /*no hastname in the url */
    fprintf(stderr, "No hostname in url: %s\n", url_orig);
    return -1;
  }
#if 0
  res->host = strndup(url, strchr(url, host_end) -url);
#else
  int hlen = (int) (strchr(url, host_end) -url);
  res->host = new char[hlen + 1];
  memcpy(res->host, url, hlen);
  res->host[hlen] = '\0';
#endif
  
  /* Get the file name. */
  url = strchr(url, path_start);
  const char * end = url + strlen(url);
  end--;
  //handle broken urls with multiple slashes
  while((end > url) && (*end == '/')) end--;
  end++;
  if((path_start == 0) || (end == url)
     || (strncmp(url, "/", end - url) ==  0)){
    res->name = strdup(res->host);
  }else{
    while(strchr(url, '/') && (strchr(url, '/') < end))
      url = strchr(url, '/') + 1;
#if 0
    res->name = strndup(url, end - url);
#else
    int nlen = (int) (end -url);
    res->name = new char[nlen + 1];
    memcpy(res->name, url, nlen);
    res->name[nlen] = '\0';
#endif
  }
  
  return res->proto;
}




/*
 * Socket operations that abstract ssl and keepalive as much as possible.
 * Keepalive is set when parsing the headers.
 *
 */

int 
XrdOssHttpRead::close_client_socket() {
  if (url->sock_type == SOCK_KEEPALIVE) return SOCK_KEEPALIVE;
  return close_client_force();
}


int
XrdOssHttpRead::close_client_force() {
  if(url->sock_type != SOCK_CLOSED){
    close(url->sockfd);
  }
  return url->sock_type = SOCK_CLOSED;
}

int 
XrdOssHttpRead::read_client_socket(void * buf, size_t len) {
  int res;
  struct timeval timeout;
  timeout.tv_sec = url->timeout;
  timeout.tv_usec = 0;
  setsockopt(url->sockfd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
  res = read(url->sockfd, buf, len);
  //  if(res <= 0) errno_report("read");
  return res;
}

ssize_t
XrdOssHttpRead::write_client_socket(const void * buf, size_t len)
{
  do {
    int fd = open_client_socket();
    ssize_t res;
    
    if(fd < 0) return -1; /*error hopefully reported by open*/
    res = write(url->sockfd, buf, len);
    //    if(res <= 0) errno_report("write");
    if ( !(res <= 0) || (url->sock_type != SOCK_KEEPALIVE )) return res;
    
    /* retry a failed keepalive socket */
    close_client_force();
  } while (url->sock_type == SOCK_KEEPALIVE);
  return -1; /*should not reach*/
}

/*
 * Function yields either a positive int after connecting to
 * host 'hostname' on port 'port'  or < 0 in case of error
 *
 * It handles keepalive by not touching keepalive sockets.
 * The SSL context is created so that read/write can use it.
 *
 * hostname is something like 'www.tmtd.de' or 192.168.0.86
 * port is expected in machine order (not net order)
 *
 * ((Flonix  defines USE_IPV6))
 *
 */
#if defined(AF_INET6) && defined(IN6_IS_ADDR_V4MAPPED)
#define USE_IPV6
#endif

int 
XrdOssHttpRead::open_client_socket() {
#ifdef USE_IPV6
  struct addrinfo hints;
  char portstr[10];
  int gaierr;
  struct addrinfo* ai;
  struct addrinfo* aiv4;
  struct addrinfo* aiv6 = 0;
  struct sockaddr_in6 sa;
#else /* USE_IPV6 */
  struct hostent *he;
  struct sockaddr_in sa;
#endif /* USE_IPV6 */
  int sa_len, sock_family, sock_type, sock_protocol;
  
  if(url->sock_type == SOCK_KEEPALIVE) return url->sock_type;
  if(url->sock_type != SOCK_CLOSED) close_client_socket();
  
  (void) memset((void*) &sa, 0, sizeof(sa));
  
#ifdef USE_IPV6
  (void) memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  (void) snprintf(portstr, sizeof(portstr), "%d", (int) url->port);
  if ((gaierr = getaddrinfo(url->host, portstr, &hints, &ai)) != 0) {
    (void) fprintf(stderr, "getaddrinfo %s - %s\n",
		   url->host, gai_strerror(gaierr));
    return -1;
  }
  
  /* Find the first IPv4 and IPv6 entries. */
  for (aiv4 = ai; aiv4 != NULL; aiv4 = aiv4->ai_next) {
    if (aiv4->ai_family == AF_INET)
      break;
    if ((aiv4->ai_family == AF_INET6) && (aiv6 == NULL))
      aiv6 = aiv4;
  }
  
  /* If there's an IPv4 address, use that, otherwise try IPv6. */
  if (aiv4 == NULL)
    aiv4 = aiv6;
  if (aiv4 == NULL) {
    (void) fprintf(stderr, "no valid address found for host %s\n",
		   url->host);
    errno = EIO;
    return -1;
  }
  if (sizeof(sa) < aiv4->ai_addrlen) {
    (void) fprintf(stderr, "%s - sockaddr too small (%lu < %lu)\n",
		   url->host, (unsigned long) sizeof(sa),
		   (unsigned long) aiv4->ai_addrlen);
    errno = EIO;
    return -1;
  }
  sock_family = aiv4->ai_family;
  sock_type = aiv4->ai_socktype;
  sock_protocol = aiv4->ai_protocol;
  sa_len = aiv4->ai_addrlen;
  (void) memmove(&sa, aiv4->ai_addr, sa_len);
  freeaddrinfo(ai);
  
#else /* USE_IPV6 */
  
  he = gethostbyname(url->host);
  if (he == NULL) {
    (void) fprintf(stderr, "unknown host - %s\n", url->host);
    errno = EIO;
    return -1;
  }
  sock_family = sa.sin_family = he->h_addrtype;
  sock_type = SOCK_STREAM;
  sock_protocol = 0;
  sa_len = sizeof(sa);
  (void) memmove(&sa.sin_addr, he->h_addr, he->h_length);
  sa.sin_port = htons(url->port);
  
#endif /* USE_IPV6 */
  
  url->sockfd = socket(sock_family, sock_type, sock_protocol);
  if (url->sockfd < 0) {
    //    errno_report("couldn't get socket");
    
    return -1;
  }
  if (connect(url->sockfd, (struct sockaddr*) &sa, sa_len) < 0) {
    //    errno_report("couldn't connect socket");
    return -1;
  }
  
#ifdef USE_SSL
  if ((url->proto) == PROTO_HTTPS) {
    /* Make SSL connection. */
    int r;
    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    url->ssl_ctx = SSL_CTX_new(SSLv23_client_method());
    url->ssl = SSL_new(url->ssl_ctx);
    SSL_set_fd(url->ssl, url->sockfd);
    r = SSL_connect(url->ssl);
    if (r <= 0) {
      close(url->sockfd);
      (void) fprintf(stderr, "%s:%d - SSL connection failed - %d\n",
		     url->host, url->port, r);
      ERR_print_errors_fp(stderr);
      errno = EIO;
      return -1;
    }
  }
#endif
  return url->sock_type = SOCK_OPEN;
}

/*
 * Scan the received header for interesting fields. Since C does not have
 * tools for working with potentially unterminated strings this is quite
 * long and ugly.
 *
 * Return the length of the header in case part of the data was
 * read with the header.
 * Content-Length means different thing whith GET and HEAD.
 */

ssize_t
XrdOssHttpRead::parse_header(const char * buf, ssize_t bytes,
        const char * method, size_t * content_length, int expect)
{
    int status;
    const char * ptr = buf;
    const char * end;
    int seen_accept = 0, seen_length = 0, seen_close = 0;

    if (bytes <= 0) {
        return -1;
    }

    end = (char*)memchr(ptr, '\n', bytes);
    if(!end) {
      //        plain_report ( "reply does not contain newline!", method, buf, 0);
        errno = EIO;
        return -1;
    }
    end = ptr;
    while(1){
        end = (char*)memchr((end + 1), '\n', bytes - (end - ptr));
        if(!end || ((end + 1) >= (ptr + bytes)) ) {
	  //            plain_report ("reply does not contain end of header!",
	  //                    method, buf, bytes);
            errno = EIO;
            return -1;
        }
        if(mempref(end, "\n\r\n", bytes - (end - ptr))) break;
    }
    size_t header_len = (end + 3) - ptr;

    end = (char*)memchr(ptr, '\n', bytes);
    const char * http = "HTTP/1.1 ";
    if(!mempref(ptr, http, end - ptr) || !isdigit( *(ptr + strlen(http))) ) {
      //        plain_report ("reply does not contain status!",
      //method, buf, header_len);
        errno = EIO;
        return -1;
    }
    status = strtol( ptr + strlen(http), (char **)&ptr, 10);
    if (status != expect) {
        fprintf(stderr, "%s: failed with status: %d%.*s.\n",
                method, status, (int)(end - ptr) - 1, ptr);
        if (!strcmp("HEAD", method)) fwrite(buf, bytes, 1, stderr); /*DEBUG*/
        errno = EIO;
        if (status == 404) errno = ENOENT;
        return -1;
    }

    const char * content_length_str = "Content-Length: ";
    const char * accept = "Accept-Ranges: bytes";
    const char * date = "Last-Modified: ";
    const char * close = "Connection: close";
    struct tm tm;
    while(1)
    {
        ptr = end+1;
        if( !(ptr < buf + (header_len - 4))){
            if(! seen_accept){
//                plain_report("server must Accept-Range: bytes",
//                        method, buf, 0);
                errno = EIO;
                return -1;
            }
            if(! seen_length){
//                plain_report("reply didn't contain Content-Length!",
//                        method, buf, 0);
                errno = EIO;
                return -1;
            }
            if(url->sock_type == SOCK_OPEN && !seen_close)
                url->sock_type = SOCK_KEEPALIVE;
            if(url->sock_type == SOCK_KEEPALIVE && seen_close)
                url->sock_type = SOCK_OPEN;
            return header_len;
        }
        end = (char*)memchr(ptr, '\n', bytes - (ptr - buf));
        if( mempref(ptr, content_length_str, end - ptr)
                && isdigit( *(ptr + strlen(content_length_str))) ){
            *content_length = atoll(ptr + strlen(content_length_str));
            seen_length = 1;
            continue;
        }
        if( mempref(ptr, accept, end - ptr) ){
            seen_accept = 1;
            continue;
        }
        if( mempref(ptr, date, end - ptr) ){
            memset(&tm, 0, sizeof(tm));
            if(!strptime(ptr + strlen(date),
                        "%n%a, %d %b %Y %T %Z", &tm)){
//                plain_report("invalid time",
//                        method, ptr + strlen(date),
//                        (end - ptr) - strlen(date)) ;
                continue;
            }
            url->last_modified = mktime(&tm);
            continue;
        }
        if( mempref(ptr, close, end - ptr) ){
            seen_close = 1;
        }
    }
}

/*
 * Send the header, and get a reply.
 * This relies on 1k reads and writes being generally atomic -
 * - they fit into a single frame. The header should fit into that
 * and we do not need partial read handling so the exchange is simple.
 * However, broken sockets have to be handled here.
 */

ssize_t
XrdOssHttpRead::exchange(char * buf, const char * method,
			 size_t * content_length, off_t start, off_t end, size_t * header_length)
{
  ssize_t res;
  ssize_t bytes;
  int range = (end > 0);
  
  /* Build request buffer, starting with the request method. */
  
  bytes = snprintf(buf, HEADER_SIZE, "%s %s HTTP/1.1\r\nHost: %s\r\n",
		   method, url->path, url->host);
  bytes += snprintf(buf + bytes, HEADER_SIZE - bytes,
		    "User-Agent: %s %s\r\n", __FILE__, VERSION);
  if (range) bytes += snprintf(buf + bytes, HEADER_SIZE - bytes,
			       "Range: bytes=%llu-%llu\r\n",
			       (unsigned long long) start, (unsigned long long) end);
  bytes += snprintf(buf + bytes, HEADER_SIZE - bytes, "\r\n");
  
  
  /* Now actually send it. */
  while(1){
    /*
     * It looks like the sockets abandoned by the server do not go away.
     * Instead of returning EPIPE they allow zero writes and zero reads. So
     * this is the place where a stale socket would be detected.
     *
     * Socket that return EAGAIN cause long delays. Reopen.
     *
     * Reset errno because reads/writes of 0 bytes are a success and are not
     * required to touch it but are handled as error below.
     *
     */
    errno = 0;
    res = write_client_socket(buf, bytes);
    if ( ((res <= 0) && ! errno) || (errno == EAGAIN) || (errno == EPIPE)) {
      //      errno_report("exchange: failed to send request, retrying"); /* DEBUG */
      close_client_force();
      continue;
    }
    if (res <= 0){
      //      errno_report("exchange: failed to send request"); /* DEBUG */
      return res;
    }
    res = read_client_socket(buf, HEADER_SIZE);
    if ( ((res <= 0) && ! errno) || (errno == EAGAIN) || (errno == EPIPE)) {
      //      errno_report("exchange: did not receive a reply, retrying"); /* DEBUG */
      close_client_force();
      continue;
    } else if (res <= 0) {
      //      errno_report("exchange: failed receving reply from server"); /* DEBUG */
      return res;
    } else break;
    /* Not reached */
  }
  bytes = res;
  
  res = parse_header(buf, bytes, method, content_length,
		     range ? 206 : 200);
  if (res <= 0){
//    plain_report("exchange: server error", method, buf, bytes);
    return res;
  }
  
  if (header_length) *header_length = res;
  
  return bytes;
}

/*
 * Function uses HEAD-HTTP-Request
 * to determine the file size
 */

ssize_t 
XrdOssHttpRead::get_stat(struct stat * stbuf) {
    char buf[HEADER_SIZE];

    if( exchange(buf, "HEAD", &(url->file_size), 0, 0, 0) < 0 )
        return -1;

    close_client_socket();

    stbuf->st_mtime = url->last_modified;
    return stbuf->st_size = url->file_size;
}

/*
 * get_data does all the magic
 * a GET-Request with Range-Header
 * allows to read arbitrary bytes
 */

ssize_t 
XrdOssHttpRead::Read(char* buffer, off_t start, size_t size)
{
    char buf[HEADER_SIZE];
    const char * b;
    ssize_t bytes;
    off_t end = start + size - 1;
    char * destination = buffer;
    size_t content_length, header_length;

    bytes = exchange(buf, "GET", &content_length,
		     start, end, &header_length);
    if(bytes <= 0) return -1;
    
    if (content_length != size) {
      size = min(content_length, size);
    }


    b = buf + header_length;

    bytes -= (b - buf);
    memcpy(destination, b, bytes);
    size -= bytes;
    destination +=bytes;
    for (; size > 0; size -= bytes, destination += bytes) {

        bytes = read_client_socket(destination, size);
        if (bytes < 0) {
	  //            errno_report("GET (read)");
            return -1;
        }
        if (bytes == 0) {
            break;
        }
    }

    close_client_socket();

    return end - start + 1 - size;
}
