#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include <XrdOssCache/XrdOssReadCache.hh>
#include <XrdOssCache/XrdOssCacheTrace.hh>
#include <XrdOss/XrdOssApi.hh>
#include <XrdOssCache/XrdOssCachePlugin.hh>

#define CHECK_BIT(var,pos) ((var) & (1<<(pos)))
#define OR_BIT_PTR(var,pos) ((*var)|=(1<<(pos)))
#define XOR_BIT_PTR(var,pos) ((*var)^=(1<<(pos)))
#define RCDEBUG 0

XrdSysMutex OssReadCacheMutex;


namespace XrdOssCachePlugin
{
  extern XrdSysError OssCacheEroute;
  extern XrdOucTrace OssCacheTrace;
  extern XrdOssSys* XrdOssSS;
  const char* tident="-NA-";
}

using namespace XrdOssCachePlugin;

XrdOssReadCacheHandle::XrdOssReadCacheHandle(const char* filename, size_t filesize, time_t modtime, unsigned long modtimensec,  XrdOssReadCache* cache) {
  EPNAME("Constructor");
  unsigned char md[SHA_DIGEST_LENGTH];
  char hexmd[(SHA_DIGEST_LENGTH*2)+1];
  SHA1((const unsigned char*) filename, strlen(filename),md);
  sprintf(hexmd,"%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x",md[0],md[1],md[2],md[3],md[4],md[5],md[6],md[7],md[8],md[9],md[10],md[11],md[12],md[13],md[14],md[15],md[16],md[17],md[18],md[19]);

  SHA1filename = hexmd;

  FileName     = filename;
  Cache        = cache;
  SyncCounter  = 0;
  
  pagehits        = 0;
  pagemiss        = 0;
  requestedbytes  = 0;
  pagedbytes      = 0;

  cfname = Cache->GetDirectory(); cfname += "/"; 
  mfname = cfname; mfname += ".#";
  hfname = cfname; hfname += ".$";

  cfname += hexmd;
  mfname += hexmd;
  hfname += hexmd;

  struct stat st;
  bool isnew=false;
  if (::stat(cfname.c_str(),&st)) {
    // new object 
    Cache->IncDiskCacheObjects();
    isnew=true;
  }
  cfFd = open(cfname.c_str(), O_RDWR | O_CREAT, S_IRWXU);
  bmFd = open(mfname.c_str(), O_RDWR | O_CREAT, S_IRWXU); 
  hfFd = open(hfname.c_str(), O_RDWR | O_CREAT, S_IRWXU); 
 
  if ( cfFd <1) {
    OssCacheEroute.Say("error:  cannot open/create ", cfname.c_str());
    ok = false;
    return;
  }
  if ( bmFd <1) {
    OssCacheEroute.Say("error:  cannot open/create ", mfname.c_str());
    ok = false;
    return;
  }
  if ( hfFd <1) {
    OssCacheEroute.Say("error:  cannot open/create ", hfname.c_str());
    ok = false;
    return;
  }

  if (Cache->GetIsPublic() && (cfFd>=0)) {fchmod(cfFd, S_IRWXU | S_IRWXG | S_IRWXO);}
  if (Cache->GetIsPublic() && (bmFd>=0)) {fchmod(bmFd, S_IRWXU | S_IRWXG | S_IRWXO);}
  if (Cache->GetIsPublic() && (hfFd>=0)) {fchmod(hfFd, S_IRWXU | S_IRWXG | S_IRWXO);}

  // check that the size of these files are consistent ...
  size_t cfsize=0;
  size_t mfsize=0;
  size_t hfsize=0;
  nRefs = 0;

  bool initialize=false;

  if ((!FileSize(cfFd,cfsize)) || (!FileSize(bmFd,mfsize)) || (!FileSize(hfFd,hfsize))) {
    TRACE(ALL, "error:  get filesize error " << cfname.c_str());
    ok = false;
    return ;
  }

  //  CacheFileLen = filesize;
  CacheFileLen    = 0;
  CacheFileOffset = 0;
  CacheFile = 0;

  MapLen = (filesize/Cache->GetPageSize()/8) +1;

  // if the file is new in the cache these files have 0 size
  if (((filesize > 0) && ((filesize != cfsize))) || (!mfsize) || (!hfsize)) {
    initialize = true;
    // the orignal differs from the cache file
    if (!isnew) 
      {TRACE(ALL,"warning:  cache discrepancy (file size differs) - dropped" << cfname.c_str());}
    else
      {TRACE(ALL," creating new cache object:=" << cfname.c_str());}
    // wipe out everything from this cache file and reset the map
    if (!Truncate(filesize,MapLen)) {
      TRACE(ALL,"error:  cannot truncate cache files porperly " << cfname.c_str());
      ok= false;
      return ;
    }
  }
  
  // map the files now!

  Map = (char*) mmap(NULL, MapLen, PROT_READ|PROT_WRITE, MAP_SHARED, bmFd,0);
  Header = (Header_t*) mmap(NULL, sizeof(Header_t),     PROT_READ|PROT_WRITE, MAP_SHARED, hfFd,0);
  
  if (Map == MAP_FAILED) {
    OssCacheEroute.Say("memory mapping of map file failed" ,mfname.c_str());
    TRACE(ALL,"error:  memory mapping of map file failed" << mfname.c_str());
    ok = false;
    return ;
  }
  
  if (Header == MAP_FAILED) {
    OssCacheEroute.Say("memory mapping of header file failed", hfname.c_str());
    TRACE(ALL,"error:  memory mapping of header file failed" << hfname.c_str());
    ok = false;
    return ;
  }

  //  if (CacheFile == MAP_FAILED) {
  //    OssCacheEroute.Say("memory mapping of cache file failed" ,cfname.c_str());
  //    TRACE(ALL,"error:  memory mapping of cache file failed" << cfname.c_str());
  //    ok = false;
  //    return ;
  //  }

  //  printf("CacheFile = %lld\n",CacheFile);
  if (Header->dirty) {
    TRACE(ALL,"warning:  cache file is dirty - dropped" << cfname.c_str());
    initialize=true;
  }

  bool initializeagain=false;
  do {
    if (initialize) {
      // new initializiation
      sprintf(Header->identifier,"%s",filename);
      Header->mtime     = modtime;
      Header->mtimensec = modtimensec;
      Header->pagesize  = Cache->GetPageSize();
      Header->filesize  = filesize;
      Header->dirty     = 0;
      // erase the cache map
      memset(Map,0,MapLen);
      msync(Map,MapLen,MS_SYNC);
      msync(Header,sizeof(Header_t),MS_SYNC);
      initializeagain = false;
    }
    
    // cross check the header information
    if (strcmp(Header->identifier, filename)) {
      TRACE(ALL,"error:  inconsistent information in cache header - wrong filename " << filename);
      initializeagain = true;
      initialize = true;
    }

    if ( (Header->mtime != modtime) || (Header->mtimensec != modtimensec) ) {
      TRACE(ALL,"error:  inconsistent information in cache header - wrong modification time " << filename);
      initializeagain = true;
      initialize = true;
    }

    if ( (Header->pagesize != Cache->GetPageSize()) ) {
      // drop the cache info for this file
      initializeagain = true;
      initialize = true;
    }

    if ( (Header->filesize  != filesize) ) {
      initializeagain = true;
      initialize = true;
    }

    if (initializeagain) {
      if (!Reset(filesize)) {
	TRACE(ALL,"error:  cannot reset cache information " << filename);
	ok = false;
	return;
      }
    }

  } while (initializeagain);
	
  ok = true;

  { char logline[1024]; sprintf(logline, ":%s(line %04d) : CacheAllocationSize %.02f [%.02f] Mb cached in %u objects",__FUNCTION__,__LINE__,1.0*Cache->GetDiskCacheSize()/1024.0/1024.0, 1.0*Cache->GetSize()/1024.0/1024.0,Cache->GetDiskCacheObjects()); TRACE(Cache,logline);}


  TRACE(Debug," returning handle for " << filename);
}

XrdOssReadCacheHandle::~XrdOssReadCacheHandle() {
  OssReadCacheMutex.UnLock();
  Cache->RemoveReadCacheHandle(FileName.c_str());

  if (Map)    munmap(Map,MapLen);
  if (Header) munmap(Header,sizeof(Header_t));
  if (CacheFile>0) {munmap(CacheFile,CacheFileLen); CacheFile=0;}
  

  if (cfFd>0) {fsync(cfFd);::close(cfFd);}
  if (bmFd>0) {fsync(bmFd);::close(bmFd);}
  if (hfFd>0) {fsync(hfFd);::close(hfFd);}
}

bool 
XrdOssReadCacheHandle::Erase() {
  EPNAME("Erase");
  TRACE(ALL,"erasing " << FileName.c_str());
  if (Map)    munmap(Map,MapLen);
  if (Header) munmap(Header,sizeof(Header_t));
  if (CacheFile>0) {munmap(CacheFile,CacheFileLen); CacheFile=0;}

  Map=0;
  Header=0;
  CacheFile=0;

  if (cfFd>0) {::close(cfFd);cfFd=0;}
  if (bmFd>0) {::close(bmFd);bmFd=0;}
  if (hfFd>0) {::close(hfFd);hfFd=0;}

  // if these fail, we cannot change it
  if ( (!unlink(cfname.c_str())) &&
       (!unlink(mfname.c_str())) &&
       (!unlink(hfname.c_str())) ) return true;
  else
    return false;
}

bool 
XrdOssReadCacheHandle::Reset(size_t filesize) {
  // truncate the cachefile to 0 and back to filesize
  if ( (ftruncate(cfFd,0)) || (ftruncate(cfFd,filesize)) ) {
    return false;
  }
  return true;
}

void
XrdOssReadCacheHandle::Sync() {
  Mutex.Lock();
  if (Map)         msync(Map,MapLen,MS_SYNC);
  if (CacheFile>0) msync(CacheFile,CacheFileLen,MS_SYNC);
  // clear dirty flag
  Header->dirty     = 0;
  if (Header)   msync(Header,sizeof(Header_t),MS_SYNC);
  if (cfFd>=0) {
    // sync the cache file
    fsync(cfFd);
  }
  Mutex.UnLock();
}

bool 
XrdOssReadCacheHandle::Truncate(size_t filesize, size_t mapsize) {
  if ( (cfFd>=0)) {if(ftruncate(cfFd,filesize))         return false;}
  if ( (bmFd>=0)) {if(ftruncate(bmFd,mapsize))          return false;}
  if ( (hfFd>=0)) {if(ftruncate(hfFd,sizeof(Header_t))) return false;}
  return true;
}



bool 
XrdOssReadCacheHandle::FileSize(int fd, size_t &fsize) {
  struct stat st;
  st.st_size=0;
  if (fstat(fd, &st)) {
    return false;
  } else {
    fsize=st.st_size;
    return true;
  }
}

bool
XrdOssReadCacheHandle::GetCached(int fd, double &percentage) {
  struct stat cstat;
  if (!fstat(fd,&cstat)) {
    size_t refsize = cstat.st_size;
    size_t allocsize = cstat.st_blocks * 512;
    if (allocsize<refsize) {
      percentage = 100.0 * allocsize / refsize;
    } else {
      percentage = 100.0;
    }
    return true;
  } else {
    return false;
  }
}

bool
XrdOssReadCacheHandle::GetCachedSize(int fd, size_t &allocsize) {
  struct stat cstat;
  if (!fstat(fd,&cstat)) {
    allocsize = cstat.st_blocks * 512;
    return true;
  } else {
    return false;
  }
}


size_t        
XrdOssReadCacheHandle::Read(off_t offset, char* buffer, size_t len, XrdOssDF* oss){
  EPNAME("Read");
  Mutex.Lock();
  int pagestart = offset/Cache->GetPageSize();
  int pagestop  = (offset+len+ (Cache->GetPageSize()/2))/Cache->GetPageSize();

  int fetchstart = -1;
  int fetchstop  = -1;

  if (!len) {
    Mutex.UnLock();
    return 0;
  }

  requestedbytes += len;
  if (RCDEBUG){ char logline[1024]; sprintf(logline, ":%s(line %04d) : Read offset:=%lld len:=%ld pagestart:=%d pagestop:=%d",__FUNCTION__,__LINE__,(long long)offset,(long)len,pagestart,pagestop);TRACE(ALL,logline); }
  //  Mutex.Lock();
  // check all needed pages


  {
    // mmap this part of the file
    off_t cstart = (long long)pagestart * (long long)Cache->GetPageSize();
    long long clen  = (pagestop-pagestart+1)* Cache->GetPageSize();
    // the last page can extend the file size, we have to make sure not to read a full page in this case!
    if (RCDEBUG) { char logline[1024]; sprintf(logline, ":%s(line %04d) : Initial parameters pos:==%lld len=%lld",__FUNCTION__,__LINE__,(long long)cstart,(long long)clen); TRACE(ALL,logline);}
    if ( (cstart +clen) > (off_t)Header->filesize) { clen = Header->filesize-cstart;}
    if (CacheFile>0) {
      // if we have more than one consumer, we avoid to map larger regions than needed - no, we don't ... let's see ...
      if ( (nRefs>1) || ((off_t)CacheFileOffset > cstart) || ( (off_t)(CacheFileOffset+CacheFileLen) < cstart) || ((int)CacheFileLen > Cache->GetMapSize()) ) {
	// remove mapping and create new
	munmap(CacheFile,CacheFileLen);
	CacheFileOffset = cstart;

	// adjust the mapping to the full map window
	if (clen < Cache->GetMapSize() ) {
	  clen = Cache->GetMapSize();
	}
	// take care not to map over the end of file
	if ( (CacheFileOffset +clen) > (off_t)Header->filesize) { clen = Header->filesize-CacheFileOffset;}

	CacheFileLen    = clen;
	if (RCDEBUG) { char logline[1024]; sprintf(logline, ":%s(line %04d) : New Mapping pos:=%lld len:=%lld",__FUNCTION__,__LINE__,(long long)CacheFileOffset,(long long)CacheFileLen);TRACE(ALL,logline); }
	CacheFile  = (char*) mmap(NULL, CacheFileLen, PROT_READ|PROT_WRITE, MAP_SHARED, cfFd,CacheFileOffset);
	if (CacheFile == MAP_FAILED) {
	  Mutex.UnLock();
	  return 0;
	}
      } else {
	if ( ((off_t)(CacheFileOffset+CacheFileLen)) < (cstart + clen) ) {
	  // extend mapping
	  if ( ((cstart-CacheFileOffset)+clen) < Cache->GetMapSize()) {
	    clen = Cache->GetMapSize();
	  } else {
	    clen += CacheFileLen;
	  }
	  if (RCDEBUG) { char logline[1024]; sprintf(logline, ":%s(line %04d) :   ReMapping pos:=%lld len:=%lld :=> pos:%lld len:=%lld",__FUNCTION__,__LINE__,(long long)CacheFileOffset,(long long)CacheFileLen, (long long)CacheFileOffset, (long long)clen);TRACE(ALL,logline); }

	  if ( (CacheFileOffset +clen) > (off_t)Header->filesize) { clen = Header->filesize-CacheFileOffset;}
#ifdef __macos__
	  munmap(CacheFile, CacheFileLen);
	  CacheFile  = (char*) mmap(NULL, clen, PROT_READ|PROT_WRITE, MAP_SHARED, cfFd,CacheFileOffset);
#else
	  CacheFile  = (char*) mremap(CacheFile,CacheFileLen,clen,MREMAP_MAYMOVE);
#endif
	  if (CacheFile!=MAP_FAILED) {
	    CacheFileLen = clen;
	  } else {
	    Mutex.UnLock();
	    return 0;
	  }
	} else {
	  if (RCDEBUG) { char logline[1024]; sprintf(logline, ":%s(line %04d) : Old Mapping pos:=%lld len:=%lld",__FUNCTION__,__LINE__,(long long)CacheFileOffset,(long long)CacheFileLen);TRACE(ALL,logline); }
	}
      }
    } else {
      // do a fresh mapping
      CacheFileOffset = cstart;

      if (clen < Cache->GetMapSize()) {
	clen = Cache->GetMapSize();
      }
      if ( (CacheFileOffset +clen) > (off_t)Header->filesize) { clen = Header->filesize-CacheFileOffset;}  
      CacheFileLen    = clen;

      if (RCDEBUG) { char logline[1024]; sprintf(logline, ":%s(line %04d) : 1st Mapping pos:=%lld len:=%lld",__FUNCTION__,__LINE__,(long long)CacheFileOffset,(long long)CacheFileLen);TRACE(ALL,logline); }
      CacheFile  = (char*) mmap(NULL, CacheFileLen, PROT_READ|PROT_WRITE, MAP_SHARED, cfFd,CacheFileOffset);
      if (CacheFile == MAP_FAILED) {
	Mutex.UnLock();
	return 0;
      }
    }    
  }  
  
  for (int i=pagestart; i<= pagestop; i++) {
    if (RCDEBUG>1) { char logline[1024]; sprintf(logline, ":%s(line %04d) : Read page:=%d fs:=%d fstop:=%d",__FUNCTION__,__LINE__,i,fetchstart,fetchstop);TRACE(ALL,logline); }
    char b  = Map[i/8];
    int  sb = i%8;
    if (CHECK_BIT(b,sb)) {
      // page is there
      pagehits++;
      if (fetchstart>-1) {
	fetchstop = i-1;
	if (!GetPages(fetchstart,fetchstop,oss)) {
	  Mutex.UnLock();
	  return 0;
	}
	fetchstart =-1;
	fetchstop  =-1;
      }
    } else {
      pagemiss++;
      pagedbytes += Cache->GetPageSize();
      if (fetchstart <0) {
	fetchstart = i;
      } 
      if (i==pagestop) {
	fetchstop = i;
	if (!GetPages(fetchstart,fetchstop,oss)) {
	  Mutex.UnLock();
	  return 0;
	}
	fetchstart=-1;
	fetchstop=-1;
      }
    }
  }

  if (!memcpy(buffer,CacheFile+offset-CacheFileOffset,len)) {
    Mutex.UnLock();
    return 0;
  } else {
    Mutex.UnLock();
    if (Header->dirty)
      return 0;
    return len;
  }
}

bool 
XrdOssReadCacheHandle::GetPages(int pagestart, int pagestop, XrdOssDF* oss) {
  EPNAME("GetPages");
  off_t cstart = (long long)pagestart * (long long)Cache->GetPageSize();
  long long clen  = (pagestop-pagestart+1)* Cache->GetPageSize();
  
  // verify if we need to cleanup
  if (Cache->GetDiskCacheObjects() > Cache->GetMaxObjects()) {
    // we need to do cleanup of objects
    Cache->CleanupSem.Post();
    // we read direct until there is space
    return false;
  }

  if (Cache->GetDiskCacheSize() > Cache->GetSize()) {
    // we need to do cleanup of space
    Cache->CleanupSem.Post();
    // we read direct until there is space
    return false;
  }

  // the last page can extend the file size, we have to make sure not to read a full page in this case!
  if ( (cstart +clen) > (off_t)Header->filesize) { clen = Header->filesize-cstart;}

  if (clen <0) 
    return true;

  if (RCDEBUG) { char logline[1024]; sprintf(logline, "%s(line %04d) : GetPages start:=%d stop:=%d offset:=%lld len:=%lld",__FUNCTION__,__LINE__,pagestart,pagestop,(long long)cstart,clen);TRACE(ALL,logline); }
  size_t nread = ((XrdOssCacheFile*)oss)->ReadNoCache(CacheFile-CacheFileOffset+cstart, cstart, clen);
  if ((long long) nread != clen) {
    return false;
  } else {
    // TODO: this evt. should have a mutex !!!
    Cache->AddDiskCacheSize(clen);
    // mark as read
    for (int i=pagestart; i<=pagestop; i++) {
      char* mapbyte = Map+(i/8);
      int mapbit  = i%8;
      if (RCDEBUG>1) { char logline[1024]; sprintf(logline, "%s(line %04d) : MarkPage page:=%d byte:=%d bit:=%d",__FUNCTION__,__LINE__,i,i/8, mapbit);TRACE(ALL,logline); }
      OR_BIT_PTR(mapbyte,mapbit);
    }
  }

  SyncCounter++;
  if (!(SyncCounter%Cache->GetSyncEvery_n_IO())) {
    if (RCDEBUG) { char logline[1024]; sprintf(logline, "%s(line %04d) : Synchronous Flush ",__FUNCTION__,__LINE__);TRACE(ALL,logline); }
    // synchronize cache
    msync(CacheFile-CacheFileOffset + cstart,clen,MS_SYNC);
    // synchronize map
    msync(Map, MapLen,MS_SYNC);
    // clear dirty flag
    Header->dirty     = 0;
    msync(Header,sizeof(Header_t),MS_SYNC);
  } else {
    if (RCDEBUG) { char logline[1024]; sprintf(logline, "%s(line %04d) : Asynchronous Flush ",__FUNCTION__,__LINE__);TRACE(ALL,logline); }
    msync(CacheFile-CacheFileOffset + cstart,clen,MS_ASYNC);
    msync(Map, MapLen,MS_ASYNC);
    // set dirty flag
    if (!Header->dirty) {
      // check if this page became dirty or has been dropped
      Header->dirty=1;
      msync(Header,sizeof(Header_t),MS_SYNC);
    }
  }


  if (Cache->GetIsSync()) {
    msync(CacheFile - CacheFileOffset + cstart,clen,MS_SYNC);
  }
  
  return true;
}


void
XrdOssReadCacheHandle::Drop(off_t offset, size_t len){
  int pagestart = offset/Cache->GetPageSize();
  int pagestop  = (offset+len+ (Cache->GetPageSize()/2))/Cache->GetPageSize();
  for (int i=pagestart; i<=pagestop; i++) {
    char* mapbyte = Map+(i/8);
    int mapbit = i%8;
    if (CHECK_BIT(*mapbyte,mapbit)) {
      XOR_BIT_PTR(mapbyte,mapbit);
    }
  }
}


XrdOssReadCache::XrdOssReadCache(const char* cachedirectory, int pagesize, long long size, unsigned int maxobjects, int syncinterval, float threshold, long long mapsize, bool ispublic) {
  EPNAME("Constructor");
  Directory  = cachedirectory;
  Size       = size;
  if (size == 0) {
    Size = (long long) 1024.0*1024*1024*8;
    char fval[1024]; sprintf(fval,"Setting Size to %lld\n", Size);
    TRACE(ALL,fval);
  }
  PageSize   = pagesize;
  MaxObjects = maxobjects;
  Threshold  = threshold;
  IsPublic   = ispublic;
  WriteSync  = false;
  MapSize    = mapsize;

  // we sync anyway every 10000 IO/s
  SyncEvery_n_IO = 10000;

  SyncInterval = syncinterval;

  DiskCacheSize    = 0;
  DiskCacheObjects = 0;

  TRACE(ALL,"=====> Initializing Ofs ReadCache *** Version " << XrdOssReadCacheVERSION);
  TRACE(ALL,"=====>  cache directory    => " << cachedirectory);
  char saysize[1024]; sprintf(saysize,"%lld",size);
  TRACE(ALL,"=====>  size               => " << saysize);
  char saypagesize[1024]; sprintf(saypagesize,"%d",pagesize);
  TRACE(ALL,"=====>  pagesize           => " << saypagesize);
  char sayobjects[1024]; sprintf(sayobjects,"%u",maxobjects);
  TRACE(ALL,"=====>  maxobjects         => " << sayobjects);
  char saythreshold[1024]; sprintf(saythreshold,"%.02f",threshold);
  TRACE(ALL,"=====>  threshold          => " << saythreshold);
  char saydir[1024]; sprintf(saydir,"%s",cachedirectory);
  TRACE(ALL,"=====>  directory          => " << saydir);
  char saysync[1024]; sprintf(saysync,"%d",syncinterval);
  TRACE(ALL,"=====>  syncinterval       => " << saysync);
  char saymapsize[1024]; sprintf(saymapsize,"%lld",mapsize);
  TRACE(ALL,"=====>  mapsize            => " << saymapsize);


  if (ispublic) {
    TRACE(ALL,"=====>  public cache");
  } else {
    TRACE(ALL,"=====> ofsReadCache: private cache");
  }

  // startup cleanup/sync thread
  int retc;
  pthread_t tid;

  if ((retc = XrdSysThread::Run(&tid, XrdOssReadCache::XrdOssReadCacheSync,
				(void *)(this))) < 0) {
    OssEroute.Emsg("XrdOssReadCache", retc, "creating OssReadCache cleanup/sync thread; "
		   "OssReadCache cleanup/sync thread failed.");
  }
}

XrdOssReadCache::~XrdOssReadCache() {
  //  EPNAME("Destructor");
}

bool 
XrdOssReadCache::Init() {
  EPNAME("Init");

  if (::access(Directory.c_str(), F_OK|W_OK|R_OK)) {
    TRACE(ALL,"error:  cannot access cache directory " << Directory.c_str());
    return false;
  }
  
  // scan the cache size on disk ... 
  DIR *dir;
  struct dirent *dp;
  
  if ((dir = ::opendir(Directory.c_str()))) {
    while ((dp = readdir (dir)) != NULL) {
      char fullpath[8192];
      sprintf(fullpath,"%s/%s",Directory.c_str(),dp->d_name);
      if ( (!strcmp(dp->d_name,".")) || (!strcmp(dp->d_name,"..")) 
	   || (!strncmp(dp->d_name,".#",2)) || (!strncmp(dp->d_name,".$",2))
	   || (!strcmp(dp->d_name,"namespace")))
	continue;
      double percentage=-1;
      size_t allocsize=0;
      int lfd;
      if ( (lfd=::open(fullpath,0))>=0) {
	XrdOssReadCacheHandle::GetCached(lfd,percentage);
	if (XrdOssReadCacheHandle::GetCachedSize(lfd,allocsize)) {
	  DiskCacheSize += allocsize;
	  DiskCacheObjects++;
	}
	close(lfd);
      }
      { char logline[1024]; sprintf(logline, "%s(line %04d) : ReadCache %s (%03.2f %% cached)",__FUNCTION__,__LINE__,dp->d_name,percentage);TRACE(ALL,logline); }

    }
    closedir(dir);
  }

  { char logline[1024]; sprintf(logline, "%s(line %04d) : CacheAllocationSize %.02f [%.02f] Mb cached in %u objects",__FUNCTION__,__LINE__,1.0*DiskCacheSize/1024.0/1024.0, 1.0*Size/1024.0/1024.0, DiskCacheObjects); TRACE(ALL,logline);}

  return true;
}

XrdOssReadCacheHandle* 
XrdOssReadCache::GetReadCacheHandle(const char* filename, size_t filesize, time_t modtime, unsigned long modtimensec) {
  //  EPNAME("GetReadCacheHandle");
  XrdOssReadCacheHandle* handle = NULL;
  OssReadCacheMutex.Lock();
  if (!(handle = Handles.Find(filename))) {
    handle = new XrdOssReadCacheHandle(filename, filesize, modtime, modtimensec, this);
    if (handle) {
      if (handle->ok) {
	Handles.Add(filename, handle);
	SHA1Handles.Add(handle->SHA1filename.c_str(),new XrdOucString(handle->SHA1filename.c_str()));
      } else {
	delete handle;
	OssReadCacheMutex.UnLock();
	return 0;
      }
    } else {
      OssReadCacheMutex.UnLock();
      return 0;
    }
  }
  handle->IncRefs();
  OssReadCacheMutex.UnLock();
  return handle;
}

void
XrdOssReadCache::RemoveReadCacheHandle(const char* filename, bool erase) {
  EPNAME("RemoveReadCacheHandle");

  OssReadCacheMutex.Lock();
  XrdOssReadCacheHandle* handle = Handles.Find(filename);

  if (handle) {
    handle->Sync();
    handle->DecRefs();

    if (erase) {
      // drop all pages
      handle->Drop(0,handle->Header->filesize);
      // free space
      handle->Truncate(0,0);
      handle->Truncate(handle->Header->filesize,handle->MapLen);
      // mark as dirty
      handle->Header->dirty=1;
    }
      
    if (handle->GetRefs()<=0) {
      TRACE(ALL," dropping handle object with last ref for " << filename);
      if (erase) {
	handle->Erase();
      }
      SHA1Handles.Del(handle->SHA1filename.c_str());
      Handles.Del(filename);
    } else {
      char line[1024]; sprintf(line,"%d",handle->GetRefs());
      TRACE(ALL," number of refs:=" <<line);
    }
  }
  OssReadCacheMutex.UnLock();
}

int compare(const void* a, const void *b) {
  struct filestat {
    struct stat buf;
    char filename[256];
  };
  return ( (((struct filestat*)a)->buf.st_mtime) - ((struct filestat*)b)->buf.st_mtime);
}

void
XrdOssReadCache::CleanUpObjects(bool byspace) {
  EPNAME("CleanUpObjects");
  // scan the cache size on disk ... 
  DIR *dir;
  struct dirent *dp;
  struct filestat {
    struct stat buf;
    char filename[256];
  };
  struct filestat* allstat;

  int nobjects=0;
  long tdp=0;

  TRACE(ALL," doing cache scrubbing ...");

  if ((dir = opendir(Directory.c_str()))) {
    tdp = telldir(dir);
    while ((dp = readdir (dir)) != NULL) {
      if ( (!strcmp(dp->d_name,".")) || (!strcmp(dp->d_name,".."))
	   || (!strncmp(dp->d_name,".#",2)) || (!strncmp(dp->d_name,".$",2 )))
	continue;
      
      nobjects++;
    }
    allstat = (struct filestat*) malloc(sizeof(struct filestat) * nobjects);
    if (!allstat) {
      TRACE(ALL,"error:  cannot allowcate sort array - abort!");
      exit(-2);
    }
    // go back
    seekdir(dir,tdp);
    int i=0;
    while ((dp = readdir (dir)) != NULL) {
      if ( (!strcmp(dp->d_name,".")) || (!strcmp(dp->d_name,".."))  
	|| (!strncmp(dp->d_name,".#",2)) || (!strncmp(dp->d_name,".$",2 ))
	|| (!strcmp(dp->d_name,"namespace")))
	continue;
      char fullpath[8192];
      sprintf(fullpath,"%s/%s",Directory.c_str(),dp->d_name);
      
      sprintf(allstat[i].filename,"%s",dp->d_name);
      if (stat(fullpath, &(allstat[i].buf))) {
	TRACE(ALL,"error:  cannot stat after readdir file:=" <<fullpath);
      }
      i++;
    }
    closedir(dir);
    // do the sorting
    qsort(allstat,nobjects, sizeof(struct filestat),compare);
    
    if (byspace) {
      for (i=0; i< nobjects;  i++) {
	char fullpath[8192];
	char headerpath[8192];
	char mappath[8192];

	if (GetInUseBySHA1name(allstat[i].filename)) {
	  TRACE(ALL,"info:   cache file in use file:=" << allstat[i].filename);
	  continue;
	}

	sprintf(fullpath,"%s/%s",Directory.c_str(),allstat[i].filename);
	sprintf(headerpath,"%s/.$%s",Directory.c_str(),allstat[i].filename);
	sprintf(mappath,"%s/.#%s",Directory.c_str(),allstat[i].filename);
	if (unlink(fullpath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" <<fullpath);
	}
	if (unlink(headerpath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" <<headerpath);
	}
	if (unlink(mappath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" <<mappath);
	}

	// this is freed space
	//	printf("%d %d \n",DiskCacheSize,allstat[i].buf.st_blocks);
	DiskCacheSize -= (allstat[i].buf.st_blocks*512);
	DiskCacheObjects--;
	if (DiskCacheSize <0) 
	  DiskCacheSize=0;

	{ char logline[1024]; sprintf(logline, "%s(line %04d) : Cleanup %s [%.02f MB => %.02f MB ]",__FUNCTION__,__LINE__,fullpath, DiskCacheSize*1.0/1024/1024, Size*1.0/1024/1024);TRACE(ALL,logline); }

	if ( DiskCacheSize < (Size * (1.0-Threshold))) {
	  // terminate if we get under the threshold
	  break;
	}
      }
    } else {
      for (i=0; i< nobjects - (MaxObjects*(1.0-Threshold)); i++) {
	char fullpath[8192];
	char headerpath[8192];
	char mappath[8192];
	if (GetInUseBySHA1name(allstat[i].filename)) {
	  TRACE(ALL,"info:   cache file in use file:=" << allstat[i].filename);
	  continue;
	}
	sprintf(fullpath,"%s/%s",Directory.c_str(),allstat[i].filename);
	sprintf(headerpath,"%s/.$%s",Directory.c_str(),allstat[i].filename);
	sprintf(mappath,"%s/.#%s",Directory.c_str(),allstat[i].filename);
	if (unlink(fullpath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" << fullpath);
	}
	if (unlink(headerpath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" << headerpath);
	}
	if (unlink(mappath)) {
	  TRACE(ALL,"error:  cannot cleanup cache file:=" << mappath);
	}

	// this is freed space
	DiskCacheSize -= (allstat[i].buf.st_blocks*512);
	DiskCacheObjects--;
	if (DiskCacheSize <0) 
	  DiskCacheSize=0;

	{ char logline[1024]; sprintf(logline, "%s(line %04d) : Cleanup %s [%d out of %d objects - limit %d]",__FUNCTION__,__LINE__,fullpath, i, nobjects, MaxObjects);TRACE(ALL,logline); }
      }
    }
    free(allstat);
  }
  { char logline[1024]; sprintf(logline, "%s(line %04d) : CacheAllocationSize %.02f [%.02f] Mb cached in %u objects",__FUNCTION__,__LINE__,1.0*DiskCacheSize/1024.0/1024.0, 1.0*Size/1024.0/1024.0, DiskCacheObjects); TRACE(ALL,logline);}
}

bool 
XrdOssReadCache::GetInUseBySHA1name(const char* name) {
  if (SHA1Handles.Find(name)) 
    return true;
  else 
    return false;
}


int
XrdOssReadCache::ApplySync(const char *key, XrdOssReadCacheHandle* handle, void *arg) {
  handle->Sync();
  return 0;
}


void* 
XrdOssReadCache::XrdOssReadCacheSync(void *myself) {
  EPNAME("ReadCacheSync");
  const char* tident = "-NA-";
  static time_t lastsync=0;
  XrdOssReadCache* Cache = (XrdOssReadCache*) myself;

  while (1) { 
    // wait that we are signaled about missing space ... or just regulary after timeout
    if (!Cache->CleanupSem.Wait(Cache->GetSyncInterval())) {
      // signaleed, remove pending signals
      while (!Cache->CleanupSem.Wait(0));
    }

    TRACE(Sync,"++++ wakeup cleanup/sync thread");
    
    // verify if we need to cleanup
    if (Cache->GetDiskCacheObjects() > Cache->GetMaxObjects()) {
      // we need to do cleanup of objects
      Cache->CleanUpObjects(false);
    }
    if (Cache->GetDiskCacheSize() > Cache->GetSize()) {
      // we need to do cleanup of space
      Cache->CleanUpObjects(true);
    }
    
    time_t now = time(NULL);
    
    if ( (now - lastsync) > Cache->GetSyncInterval()) {
      // ok sync everything in the cache
      TRACE(Sync,"++++ running sync ...");
      lastsync = now -1;
      OssReadCacheMutex.Lock(); 
      Cache->Handles.Apply(XrdOssReadCache::ApplySync,NULL);
      OssReadCacheMutex.UnLock();
    }
  }
  return NULL;
}
