#ifndef _XRDOSSCACHE_H
#define _XRDOSSCACHE_H

#include <XrdOuc/XrdOucString.hh>
#include <XrdSys/XrdSysPthread.hh>
#include <XrdSys/XrdSysSemWait.hh>
#include <XrdSys/XrdSysError.hh>
#include <XrdOuc/XrdOucHash.hh>
#include <XrdOss/XrdOss.hh>
#include <openssl/sha.h>

#define XrdOssReadCacheVERSION "1.0.0"
extern XrdSysError OfsEroute;
extern XrdSysMutex OssReadCacheMutex;

class XrdOssReadCache;
class XrdOssReadCacheHandle;

class XrdOssReadCacheHandle {
private:

  XrdOssReadCache* Cache;

public:
  XrdOucString SHA1filename;
  XrdOucString FileName;
  XrdOucString cfname;
  XrdOucString mfname;
  XrdOucString hfname;

  long long    pagehits;
  long long    pagemiss;
  long long    requestedbytes;
  long long    pagedbytes;

  struct CacheHeader {
    char      identifier[4096];
    time_t    mtime;
    unsigned  long mtimensec;
    int       pagesize;
    long long cachesize;
    size_t    filesize;
    int       dirty;
  };

  typedef struct CacheHeader Header_t;


  int cfFd; // cache file fd
  int bmFd; // cache bitmap file
  int hfFd; // header file fd

  int nRefs; // number of references to this object
  int SyncCounter; // counter to switch between async & sync page flushing

  static bool   FileSize(int fd, size_t &fsize);
  static bool   GetCached(int fd, double &percentage);
  static bool   GetCachedSize(int fd, size_t &allocsize);
  bool          Truncate(size_t filesize, size_t mapsize);
  bool          Reset(size_t filesize);
  bool          ok;
  Header_t*     Header;
  char*         Map;
  char*         CacheFile;
  size_t        CacheFileLen;
  off_t         CacheFileOffset;
  size_t        MapLen;
  
  void          IncRefs() {nRefs++;}
  void          DecRefs() {nRefs--;}
  int           GetRefs() {return nRefs;}

  size_t        Read(off_t offset, char* buffer, size_t len, XrdOssDF* oss);
  void          Drop(off_t offset, size_t len);
  bool          GetPages(int pagestart, int pagestop, XrdOssDF* oss);
  void          Sync();
  XrdSysMutex   Mutex;
  XrdSysError*   eDest;
  bool          Erase();

  XrdOssReadCacheHandle(const char* filename, size_t filesize, time_t modtime, unsigned long modtimensec, XrdOssReadCache* cache);
  ~XrdOssReadCacheHandle();
};



class XrdOssReadCache {
private:
  XrdOucString Directory;
  long long    Size;
  int          PageSize;
  int          MaxObjects;
  bool         IsPublic;
  bool         WriteSync;
  float        Threshold;
  long long    DiskCacheSize;
  int          DiskCacheObjects;
  int          SyncEvery_n_IO;
  int          SyncInterval;
  long long    MapSize;
  
public:

  XrdOucHash<XrdOssReadCacheHandle> Handles;
  XrdOucHash<XrdOucString> SHA1Handles;

  bool Init();
  bool SetSync(bool setting) { WriteSync = setting; }
  void RemoveReadCacheHandle(const char* filename, bool erase=false);

  XrdSysSemWait CleanupSem;

  XrdOssReadCacheHandle* GetReadCacheHandle(const char* filename, size_t filesize, time_t modtime, unsigned long modtimensec);

  XrdOssReadCache(const char* cachedirectory="/pool/ofscache", int pagesize=4*1024, long long size=0 /*8G*/, unsigned int maxobjects=1024, int syncinterval = 10, float threshold = 0.1, long long mapsize = 1024*1024l, bool ispublic=true);
  ~XrdOssReadCache();
  const char*  GetDirectory()        {return Directory.c_str();}
  bool         GetIsPublic()         {return IsPublic;}
  bool         GetIsSync()           {return WriteSync;}
  int          GetPageSize()         {return PageSize;}
  long long    GetSize()             {return Size;}  
  long long    GetDiskCacheSize()    {return DiskCacheSize;}
  long long    GetMapSize()          {return MapSize;} 
  int          GetMaxObjects()       {return MaxObjects;}
  int          GetDiskCacheObjects() {return DiskCacheObjects;}
  int          GetSyncEvery_n_IO()   {return SyncEvery_n_IO;}
  int          GetSyncInterval()     {return SyncInterval;}
  void         IncDiskCacheObjects() {DiskCacheObjects++;}
  void         AddDiskCacheSize(long long size) {DiskCacheSize += size;}
  void         CleanUpObjects(bool byspace); 
  bool         GetInUseBySHA1name(const char* sha1name);

  static void* XrdOssReadCacheSync(void *myself);
  static int   ApplySync(const char *key, XrdOssReadCacheHandle* handle, void *arg);
};

#endif

